#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generator function (with yield) related templates."""

template_generator_context_maker_decl = """\
static PyObject *%(generator_maker_identifier)s(%(generator_creation_args)s);
"""

template_generator_context_body_template = """
#if %(has_heap_declaration)s
struct %(function_identifier)s_locals {
%(function_local_types)s
};
#endif

static PyObject *%(function_identifier)s_context(PyThreadState *tstate, struct Nuitka_GeneratorObject *generator, PyObject *yield_return_value) {
    CHECK_OBJECT(generator);
    assert(Nuitka_Generator_Check((PyObject *)generator));
    CHECK_OBJECT_X(yield_return_value);

#if %(has_heap_declaration)s
    // Heap access.
%(heap_declaration)s
#endif

    // Dispatch to yield based on return label index:
%(function_dispatch)s

    // Local variable initialization
%(function_var_inits)s

    // Actual generator function body.
%(function_body)s

%(generator_exit)s
}

static PyObject *%(generator_maker_identifier)s(%(generator_creation_args)s) {
    return Nuitka_Generator_New(
        %(function_identifier)s_context,
        %(generator_module)s,
        %(generator_name_obj)s,
#if PYTHON_VERSION >= 0x350
        %(generator_qualname_obj)s,
#endif
        %(code_identifier)s,
        %(closure_name)s,
        %(closure_count)d,
#if %(has_heap_declaration)s
        sizeof(struct %(function_identifier)s_locals)
#else
        0
#endif
    );
}
"""

template_make_generator = """\
%(closure_copy)s
%(to_name)s = %(generator_maker_identifier)s(%(args)s);
"""

template_make_empty_generator = """\
%(closure_copy)s
%(to_name)s = Nuitka_Generator_NewEmpty(
    %(generator_module)s,
    %(generator_name_obj)s,
#if PYTHON_VERSION >= 0x350
    %(generator_qualname_obj)s,
#endif
    %(code_identifier)s,
    %(closure_name)s,
    %(closure_count)d
);
"""


template_generator_exception_exit = """\
%(function_cleanup)s
    return NULL;

    function_exception_exit:
%(function_cleanup)s
    CHECK_EXCEPTION_STATE(&%(exception_state_name)s);
    RESTORE_ERROR_OCCURRED_STATE(tstate, &%(exception_state_name)s);

    return NULL;
"""

template_generator_no_exception_exit = """\
    // Return statement need not be present.
%(function_cleanup)s
    return NULL;
"""

template_generator_return_exit = """\
    NUITKA_CANNOT_GET_HERE("Generator must have exited already.");
    return NULL;

    function_return_exit:
#if PYTHON_VERSION >= 0x300
    generator->m_returned = %(return_value)s;
#endif

%(function_cleanup)s
    return NULL;
"""


from . import TemplateDebugWrapper  # isort:skip

TemplateDebugWrapper.checkDebug(globals())

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Download utilities and extract locally when allowed.

Mostly used on Windows, for dependency walker and ccache binaries.
"""

import os

from nuitka.Progress import withNuitkaDownloadProgressBar
from nuitka.PythonVersions import python_version
from nuitka.Tracing import general
from nuitka.Version import getNuitkaVersion

from .AppDirs import getCacheDir
from .FileOperations import (
    addFileExecutablePermission,
    deleteFile,
    getNormalizedPath,
    getNormalizedPathJoin,
    makePath,
    queryUser,
)


def getCertifiModule():
    try:
        # Try to import vendored certifi first
        from pip._vendor import certifi
    except ImportError:
        try:
            # Fallback to system-installed certifi
            import certifi  # type: ignore
        except ImportError:
            certifi = None  # type: ignore

    return certifi


def shouldDownload(message, reject_message, assume_yes_for_downloads, download_ok):
    """Check if we should download"""

    if not download_ok:
        reply = "no"
    elif assume_yes_for_downloads:
        reply = "yes"
    else:
        reply = queryUser(
            question="""\
%s

Is it OK to download and put it in local user cache.

Fully automatic, cached. Proceed and download"""
            % (message,),
            choices=("yes", "no"),
            default="yes",
            default_non_interactive="no",
        )

    if reply != "yes":
        if reject_message is not None:
            if not download_ok:
                reject_message += " Make sure to allow downloading it when prompted."

            return general.sysexit(reject_message)

        return False

    return True


def getDownload(name, url, download_path):
    if python_version < 0x300:
        from urllib2 import (  # pylint: disable=I0021,import-error
            Request,
            urlopen,
        )
    else:
        from urllib.request import Request, urlopen

    ssl_context = None

    certifi = getCertifiModule()

    if certifi is not None:
        certifi_path = certifi.where()

        if certifi_path and os.path.exists(certifi_path):
            import ssl

            # spell-checker: ignore cafile
            ssl_context = ssl.create_default_context(cafile=certifi_path)

    # Be fair and provide a user agent
    request_headers = {"User-Agent": "Nuitka Downloader/%s" % getNuitkaVersion()}

    def _performDownload(current_url, current_ssl_context, report_hook):
        req = Request(current_url, headers=request_headers)

        with urlopen(req, context=current_ssl_context, timeout=30.0) as response:
            # Try to get the total size from headers
            if python_version < 0x300:
                total_size_header = response.info().getheader("Content-Length")
            else:
                total_size_header = response.getheader("Content-Length")

            total_size = (
                int(total_size_header)
                if total_size_header and total_size_header.isdigit()
                else -1
            )
            block_size_for_reading = 8192

            # Initial call for progress bar to set total size.
            if report_hook is not None:
                report_hook(0, 1, total_size)

            copied_size = 0

            # Actual file writing
            with open(download_path, "wb") as out_file:
                while True:
                    chunk = response.read(block_size_for_reading)
                    if not chunk:
                        break
                    out_file.write(chunk)
                    copied_size += len(chunk)

                    if report_hook is not None:
                        report_hook(copied_size, 1, total_size)

    with withNuitkaDownloadProgressBar(
        desc="Download %s" % name
    ) as report_hook_fn_outer:
        try:
            try:
                _performDownload(
                    url,
                    ssl_context,
                    report_hook_fn_outer,
                )
            except Exception:  # pylint: disable=broad-except
                if url.startswith("https://"):
                    http_url = "http://" + url[len("https://") :]

                    _performDownload(
                        http_url,
                        None,
                        report_hook_fn_outer,
                    )
        except BaseException:
            deleteFile(download_path, must_exist=False)
            raise


def getDownloadCacheName():
    return "downloads"


def getDownloadCacheDir():
    return getCacheDir(getDownloadCacheName())


def getCachedDownload(
    name,
    url,
    binary,
    unzip,
    flatten,
    is_arch_specific,
    specificity,
    message,
    reject,
    assume_yes_for_downloads,
    download_ok,
):
    # Many branches to deal with.
    # pylint: disable=too-many-branches,too-many-locals,too-many-return-statements

    nuitka_download_dir = getDownloadCacheDir()

    nuitka_download_dir = getNormalizedPathJoin(
        nuitka_download_dir, os.path.basename(binary).replace(".exe", "")
    )

    if is_arch_specific:
        nuitka_download_dir = getNormalizedPathJoin(
            nuitka_download_dir, is_arch_specific
        )

    if specificity:
        nuitka_download_dir = getNormalizedPathJoin(nuitka_download_dir, specificity)

    download_path = getNormalizedPathJoin(nuitka_download_dir, os.path.basename(url))
    exe_path = getNormalizedPathJoin(nuitka_download_dir, binary)

    makePath(nuitka_download_dir)

    if not os.path.isfile(download_path) and not os.path.isfile(exe_path):
        if not shouldDownload(
            message=message,
            reject_message=reject,
            assume_yes_for_downloads=assume_yes_for_downloads,
            download_ok=download_ok,
        ):
            return None

        general.info("Downloading '%s'." % url)

        try:
            getDownload(
                name=name,
                url=url,
                download_path=download_path,
            )
        except Exception as e:  # Any kind of error, pylint: disable=broad-except
            return general.sysexit(
                "Failed to download '%s' due to '%s'. Contents should manually be copied to '%s'."
                % (url, e, download_path)
            )

    if unzip:
        if not os.path.isfile(exe_path) and os.path.isfile(download_path):
            general.info("Extracting to '%s'" % exe_path)

            import zipfile

            try:
                # Not all Python versions support using it as a context manager, pylint: disable=consider-using-with
                zip_file = zipfile.ZipFile(download_path)

                for zip_info in zip_file.infolist():
                    if zip_info.filename[-1] == "/":
                        continue

                    if flatten:
                        zip_info.filename = os.path.basename(zip_info.filename)

                    zip_file.extract(zip_info, nuitka_download_dir)

            except (
                # Catching anything zip throws, pylint: disable=broad-except
                Exception
            ):
                general.info("Problem with the downloaded zip file, deleting it.")

                deleteFile(binary, must_exist=False)
                deleteFile(download_path, must_exist=True)

                return general.sysexit(
                    "Error, need '%s' as extracted from '%s'." % (binary, url)
                )

        # Check success here, and make sure it's executable.
        if os.path.isfile(exe_path):
            addFileExecutablePermission(exe_path)
        else:
            if reject:
                return general.sysexit(reject)

            return None

        return getNormalizedPath(exe_path)
    else:
        return getNormalizedPath(download_path)


def getCachedDownloadedMinGW64(
    target_arch, assume_yes_for_downloads, download_ok, experimental=False
):
    # Large URLs, pylint: disable=line-too-long
    if target_arch == "x86_64":
        if experimental:
            url = "https://github.com/brechtsanders/winlibs_mingw/releases/download/15.1.0posix-13.0.0-msvcrt-r4/winlibs-x86_64-posix-seh-gcc-15.1.0-mingw-w64msvcrt-13.0.0-r4.zip"
        else:
            url = "https://github.com/brechtsanders/winlibs_mingw/releases/download/14.2.0posix-19.1.1-12.0.0-msvcrt-r2/winlibs-x86_64-posix-seh-gcc-14.2.0-llvm-19.1.1-mingw-w64msvcrt-12.0.0-r2.zip"

        binary = r"mingw64\bin\gcc.exe"
    elif target_arch == "x86":
        if experimental:
            url = "https://github.com/brechtsanders/winlibs_mingw/releases/download/15.1.0posix-13.0.0-msvcrt-r4/winlibs-i686-posix-dwarf-gcc-15.1.0-mingw-w64msvcrt-13.0.0-r4.zip"
        else:
            url = "https://github.com/brechtsanders/winlibs_mingw/releases/download/14.2.0posix-19.1.1-12.0.0-msvcrt-r2/winlibs-i686-posix-dwarf-gcc-14.2.0-llvm-19.1.1-mingw-w64msvcrt-12.0.0-r2.zip"
        binary = r"mingw32\bin\gcc.exe"
    elif target_arch == "arm64":
        return None
    else:
        assert False, target_arch

    gcc_binary = getCachedDownload(
        name="mingw64",
        url=url,
        is_arch_specific=target_arch,
        specificity=url.rsplit("/", 2)[1],
        binary=binary,
        unzip=True,
        flatten=False,
        message="Nuitka will use gcc from MinGW64 of winlibs to compile on Windows.",
        reject="Only this specific gcc is supported with Nuitka.",
        assume_yes_for_downloads=assume_yes_for_downloads,
        download_ok=download_ok,
    )

    return gcc_binary


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Directories and paths to for output of Nuitka.

There are two major outputs, the build directory *.build and for
standalone mode, the *.dist folder.

A bunch of functions here are supposed to get path resolution from
this.
"""

import os

from nuitka.options.Options import (
    getOutputFilename,
    getOutputFolderName,
    getOutputPath,
    getPgoExecutable,
    isOnefileMode,
    isStandaloneMode,
    shallCreateAppBundle,
    shallCreateScriptFileForExecution,
    shallMakeDll,
    shallMakeModule,
)
from nuitka.utils.FileOperations import (
    addFilenameExtension,
    changeFilenameExtension,
    getNormalizedPath,
    getNormalizedPathJoin,
    hasFilenameExtension,
    isFilesystemEncodable,
    makePath,
    putTextFileContents,
    removeDirectory,
    resetDirectory,
)
from nuitka.utils.Importing import getExtensionModuleSuffix
from nuitka.utils.SharedLibraries import getDllSuffix
from nuitka.utils.Utils import isWin32OrPosixWindows, isWin32Windows

_main_module = None


def _getResultBaseName(suffix):
    # If the user specified a build basename, we use that.
    build_basename = getOutputFolderName()
    if build_basename:
        return build_basename + suffix

    return os.path.basename(
        getTreeFilenameWithSuffix(module=_main_module, suffix=suffix)
    )


def setMainModule(main_module):
    """Call this before using other methods of this module."""
    # Technically required.
    assert main_module.isCompiledPythonModule()

    # Singleton and to avoid passing this one all the time, pylint: disable=global-statement
    global _main_module
    _main_module = main_module


def hasMainModule():
    """For reporting to check if there is anything to talk about."""
    return _main_module is not None


def getMainModule():
    return _main_module


def getSourceDirectoryPath(onefile, create):
    """Return path inside the build directory."""

    # Distinct build folders for onefile mode.
    if onefile:
        suffix = ".onefile-build"
    else:
        suffix = ".build"

    filename = _getResultBaseName(suffix=suffix)

    if isWin32Windows() and not isFilesystemEncodable(filename=filename):
        filename = "_nuitka_temp" + suffix

    result = getOutputPath(path=filename)

    if create:
        makePath(path=result)

        git_ignore_filename = getNormalizedPathJoin(result, ".gitignore")

        if not os.path.exists(git_ignore_filename):
            putTextFileContents(filename=git_ignore_filename, contents="*")

    return result


def _getStandaloneDistSuffix(bundle):
    """Suffix to use for standalone distribution folder."""

    if bundle and shallCreateAppBundle() and not isOnefileMode():
        return ".app"
    else:
        return ".dist"


def _getActualOutputFolderName(bundle):
    dist_folder_name = getOutputFolderName()

    if dist_folder_name is None:
        dist_folder_name = _getResultBaseName(
            suffix=_getStandaloneDistSuffix(bundle=bundle)
        )
    else:
        # Add the suffix if not provided by the user
        standalone_dist_suffix = _getStandaloneDistSuffix(bundle=bundle)

        if not hasFilenameExtension(
            path=dist_folder_name, extensions=standalone_dist_suffix
        ):
            dist_folder_name = changeFilenameExtension(
                path=dist_folder_name, extension=standalone_dist_suffix
            )

    return dist_folder_name


def needsStandaloneDirectoryWorkaround():
    return isWin32Windows() and not isFilesystemEncodable(
        filename=_getActualOutputFolderName(bundle=False)
    )


def getStandaloneDirectoryPath(bundle, real):
    assert isStandaloneMode()

    if needsStandaloneDirectoryWorkaround() and not real:
        dist_folder_name = "_nuitka_temp"
    else:
        dist_folder_name = _getActualOutputFolderName(bundle=bundle)

    result = getOutputPath(path=dist_folder_name)

    if bundle and shallCreateAppBundle() and not isOnefileMode():
        result = getNormalizedPathJoin(result, "Contents", "MacOS")

    return result


def initStandaloneDirectory(logger):
    """Reset the standalone directory, if it exists from a previous run."""
    standalone_dir = getStandaloneDirectoryPath(bundle=False, real=False)
    standalone_dir_real = getStandaloneDirectoryPath(bundle=False, real=True)

    if shallCreateAppBundle():
        resetDirectory(
            path=changeFilenameExtension(standalone_dir_real, ".app"),
            logger=logger,
            ignore_errors=True,
            extra_recommendation="Stop previous binary.",
        )

    resetDirectory(
        path=standalone_dir,
        logger=logger,
        ignore_errors=True,
        extra_recommendation="Stop previous binary.",
    )

    if standalone_dir != standalone_dir_real:
        removeDirectory(
            path=standalone_dir_real,
            logger=logger,
            ignore_errors=True,
            extra_recommendation="Stop previous binary.",
        )


def renameStandaloneDirectory(dist_dir):
    if needsStandaloneDirectoryWorkaround():
        dist_dir_real = getStandaloneDirectoryPath(bundle=True, real=True)

        os.rename(dist_dir, dist_dir_real)
        dist_dir = dist_dir_real

    return dist_dir


def getResultBasePath(onefile=False, real=False):
    file_path = _getResultBaseName(suffix="")

    if isOnefileMode() and onefile:
        if shallCreateAppBundle():
            file_path = getNormalizedPathJoin(
                file_path + ".app", "Contents", "MacOS", file_path
            )

        return getOutputPath(path=file_path)
    elif isStandaloneMode() and not onefile:
        return getNormalizedPathJoin(
            getStandaloneDirectoryPath(bundle=True, real=real),
            file_path,
        )
    else:
        return getOutputPath(path=file_path)


def getResultFullpath(onefile, real):
    """Get the final output binary result full path."""

    result = getResultBasePath(onefile=onefile, real=real)

    if shallMakeModule():
        result += getExtensionModuleSuffix(preferred=True)
    elif shallMakeDll() and not onefile:
        # TODO: Could actually respect getOutputFilename() for DLLs, these don't
        # need to be named in a specific way.
        result += getDllSuffix()
    else:
        output_filename = getOutputFilename()

        if isOnefileMode() and output_filename is not None:
            if onefile:
                result = getOutputPath(path=output_filename)
            else:
                result = getNormalizedPathJoin(
                    getStandaloneDirectoryPath(bundle=True, real=real),
                    os.path.basename(output_filename),
                )
        elif isStandaloneMode() and output_filename is not None:
            result = getNormalizedPathJoin(
                getStandaloneDirectoryPath(bundle=True, real=real),
                os.path.basename(output_filename),
            )
        elif output_filename is not None:
            result = getOutputPath(path=output_filename)
        elif not isWin32OrPosixWindows() and not shallCreateAppBundle():
            result = addFilenameExtension(path=result, extension=".bin")

        if isWin32OrPosixWindows():
            result = addFilenameExtension(path=result, extension=".exe")

        if not isWin32OrPosixWindows() and isOnefileMode() and not onefile:
            result = addFilenameExtension(path=result, extension=".bin")

    return getNormalizedPath(path=result)


def getResultRunFilename(onefile):
    result = getResultFullpath(onefile=onefile, real=True)

    if shallCreateScriptFileForExecution():
        result = changeFilenameExtension(
            path=result,
            extension=".cmd" if isWin32Windows() else ".sh",
        )

    return result


def getTreeFilenameWithSuffix(module, suffix):
    return module.getOutputFilename() + suffix


def getPgoRunExecutable():
    return getPgoExecutable() or getResultRunFilename(onefile=False)


def getPgoRunInputFilename():
    return getPgoRunExecutable() + ".nuitka-pgo"


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Basics for Nuitka tools."""

import os
import sys


def goHome():
    """Go its own directory, to have it easy with path knowledge."""
    os.chdir(getHomePath())


my_abs_path = os.path.abspath(__file__)


def getHomePath():
    return os.path.normpath(os.path.join(os.path.dirname(my_abs_path), "..", ".."))


def setupPATH():
    """Make sure installed tools are in PATH.

    For Windows, add this to the PATH, so pip installed PyLint will be found
    near the Python executing this script.
    """
    os.environ["PATH"] = (
        os.environ["PATH"]
        + os.pathsep
        + os.path.join(os.path.dirname(sys.executable), "scripts")
    )


def addPYTHONPATH(path):
    python_path = os.getenv("PYTHONPATH", "")
    os.environ["PYTHONPATH"] = os.pathsep.join(python_path.split(os.pathsep) + [path])


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



import ast
import sys
import typing

if sys.version_info < (3, 9):
    from astunparse import unparse

    ast.unparse = unparse


class MainBlockRemover(ast.NodeTransformer):
    """AST transformer that removes 'if __name__ == "__main__":' blocks."""

    def visit_If(self, node):
        if (
            isinstance(node.test, ast.Compare)
            and isinstance(node.test.left, ast.Name)
            and node.test.left.id == "__name__"
        ):
            if (
                len(node.test.ops) == 1
                and isinstance(node.test.ops[0], ast.Eq)
                and len(node.test.comparators) == 1
            ):
                comparator = node.test.comparators[0]
                if (
                    sys.version_info >= (3, 8)
                    and isinstance(comparator, ast.Constant)
                    and comparator.value == "__main__"
                ) or (
                    sys.version_info < (3, 8)
                    and isinstance(comparator, ast.Str)
                    and comparator.s == "__main__"
                ):
                    return None
        return self.generic_visit(node)


def preprocess_source(source_code):
    tree = ast.parse(source_code)
    transformer = MainBlockRemover()
    transformed_tree = transformer.visit(tree)
    ast.fix_missing_locations(transformed_tree)
    return ast.unparse(transformed_tree)


def generate_stub_from_source(source_code, output_file_path, text_only=False):
    preprocessed_code = preprocess_source(source_code)
    tree = ast.parse(preprocessed_code)

    class StubGenerator(ast.NodeVisitor):
        def __init__(self):
            self.stubs = []
            self.imports_helper_dict = {}
            self.imports_output = set()
            self.typing_imports = typing.__all__
            self.in_class = False
            self.indentation_level = 0
            self.typevars = set()

        def visit_Import(self, node):
            for alias in node.names:
                self.imports_output.add("import %s" % alias.name)

        def visit_ImportFrom(self, node):
            module = node.module if node.module is not None else "."
            for alias in node.names:
                name = alias.name
                if module:
                    if module not in self.imports_helper_dict:
                        self.imports_helper_dict[module] = set()
                    self.imports_helper_dict[module].add(name)

        def visit_FunctionDef(self, node):
            if self.in_class:
                self.visit_MethodDef(node)
            else:
                for parent_node in ast.walk(tree):
                    if isinstance(parent_node, ast.ClassDef):
                        for child_node in parent_node.body:
                            if (
                                isinstance(child_node, ast.FunctionDef)
                                and child_node.name == node.name
                                and (child_node.lineno == node.lineno)
                            ):
                                return
                self.visit_RegularFunctionDef(node)

        def visit_Assign(self, node):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    target_name = target.id
                    target_type = ast.unparse(node.value).strip()
                    if (
                        isinstance(node.value, ast.Call)
                        and isinstance(node.value.func, ast.Name)
                        and (node.value.func.id == "TypeVar")
                    ):
                        if "typing" not in self.imports_helper_dict:
                            self.imports_helper_dict["typing"] = set()
                        self.imports_helper_dict["typing"].add("TypeVar")
                        self.typevars.add(target_name)
                        if node.value.args:
                            typevar_def = "%s = TypeVar(%s)" % (target_name, ', '.join([ast.unparse(arg) for arg in node.value.args]))
                            self.stubs.append(typevar_def + "\n\n")
                        else:
                            self.stubs.append(
                                '"%s" = TypeVar("%s")\n\n' % (target_name, target_name)
                            )
                        continue
                    if target_type in self.typing_imports:
                        self.imports_output.add(target_type)
                    if target_type in self.typing_imports:
                        stub = "%s: %s\n" % (target_name, target_type)
                        self.stubs.append(stub)
                    elif isinstance(node.value, ast.Call):
                        if isinstance(node.value.func, ast.Name):
                            if node.value.func.id == "frozenset":
                                stub = "%s = frozenset(%s)\n" % (target_name, ', '.join([ast.unparse(arg).strip() for arg in node.value.args]))
                                self.stubs.append(stub)
                            elif node.value.func.id == "namedtuple":
                                tuple_name = ast.unparse(node.value.args[0]).strip()
                                stub = "%s =  namedtuple(%s, %s)\n" % (target_name, tuple_name, ', '.join([ast.unparse(arg).strip() for arg in node.value.args[1:]]))
                                self.stubs.append(stub)
                            elif node.value.func.id == "TypeVar":
                                if "typing" not in self.imports_helper_dict:
                                    self.imports_helper_dict["typing"] = set()
                                self.imports_helper_dict["typing"].add("TypeVar")
                                self.typevars.add(target_name)
                    elif isinstance(node.value, ast.Subscript):
                        if isinstance(node.value.value, ast.Name):
                            target_name = target.id
                        target_type = ast.unparse(node.value).strip()
                        if "typing_extensions" not in self.imports_helper_dict:
                            self.imports_helper_dict["typing_extensions"] = set()
                        self.imports_helper_dict["typing_extensions"].add("TypeAlias")
                        stub = "%s: TypeAlias = %s\n" % (target_name, target_type)
                        self.stubs.append(stub)
                    elif not self.in_class:
                        stub = "%s = %s\n" % (target_name, target_type)
                        self.stubs.append(stub)
                elif isinstance(target, ast.Subscript):
                    if isinstance(target.value, ast.Name):
                        target_name = target.value.id
                    else:
                        continue
                    target_type = ast.unparse(node.value).strip()
                    stub = "%s: %s\n" % (target_name, target_type)
                    self.stubs.append(stub)

        def visit_MethodDef(self, node):
            args_list = []
            for arg in node.args.args:
                arg_type = self.get_arg_type(arg)
                args_list.append("%s: %s" % (arg.arg, arg_type))
            if node.returns:
                return_type = self.get_return_type(node.returns)
            else:
                return_type = "Any"
                if "typing" not in self.imports_helper_dict:
                    self.imports_helper_dict["typing"] = set()
                self.imports_helper_dict["typing"].add("Any")
            indent = "    " * (self.indentation_level + 1)
            if node.name == "__init__":
                return_type = "None"
            if node.decorator_list:
                for decorator in node.decorator_list:
                    if isinstance(decorator, ast.Name):
                        if decorator.id == "classmethod":
                            args_list = args_list[1:]
                            stub = "%s@classmethod\n%sdef %s(cls, %s) -> %s: ...\n" % (indent, indent, node.name, ', '.join(args_list), return_type)
                            self.stubs.append(stub)
                            return
                        elif decorator.id == "staticmethod":
                            stub = "%s@staticmethod\n%sdef %s(%s) -> %s: ...\n" % (indent, indent, node.name, ', '.join(args_list), return_type)
                            self.stubs.append(stub)
                            return
                        else:
                            stub = "%sdef %s(%s) -> %s: ...\n" % (indent, node.name, ', '.join(args_list), return_type)
                            self.stubs.append(stub)
                            return
            stub = "%sdef %s(%s) -> %s: ...\n" % (indent, node.name, ', '.join(args_list), return_type)
            self.stubs.append(stub)

        def visit_RegularFunctionDef(self, node):
            args_list = []
            for arg in node.args.args:
                arg_type = self.get_arg_type(arg)
                args_list.append("%s: %s" % (arg.arg, arg_type))
            if node.returns:
                return_type = self.get_return_type(node.returns)
            else:
                return_type = "Any"
                if "typing" not in self.imports_helper_dict:
                    self.imports_helper_dict["typing"] = set()
                self.imports_helper_dict["typing"].add("Any")
            stub = (
                "def %s(%s) -> %s:\n    ...\n" % (node.name, ', '.join(args_list), return_type)
            )
            self.stubs.append(stub)
            self.stubs.append("\n")

        def visit_ClassDef(self, node):
            previous_in_class = self.in_class
            self.in_class = True
            previous_indent = self.indentation_level
            if previous_in_class:
                self.indentation_level += 1
            class_name = node.name
            indent = "    " * self.indentation_level
            stub = ""
            case = self.special_cases(node)
            class_has_generic = False
            generic_types = []
            for base in node.bases:
                if (
                    isinstance(base, ast.Subscript)
                    and isinstance(base.value, ast.Name)
                    and (base.value.id == "Generic")
                ):
                    class_has_generic = True
                    if "typing" not in self.imports_helper_dict:
                        self.imports_helper_dict["typing"] = set()
                    self.imports_helper_dict["typing"].add("Generic")
                    if isinstance(base.slice, ast.Tuple):
                        for elt in base.slice.elts:
                            if isinstance(elt, ast.Name):
                                generic_types.append(elt.id)
                    elif isinstance(base.slice, ast.Name):
                        generic_types.append(base.slice.id)
                    for type_name in generic_types:
                        if type_name in self.typevars:
                            continue
                        self.typevars.add(type_name)
                        if "typing" not in self.imports_helper_dict:
                            self.imports_helper_dict["typing"] = set()
                        self.imports_helper_dict["typing"].add("TypeVar")
            if case == "TypedDict":
                stub = "%sclass %s(TypedDict):\n" % (indent, class_name)
                if "typing" not in self.imports_helper_dict:
                    self.imports_helper_dict["typing"] = set()
                self.imports_helper_dict["typing"].add("TypedDict")
                for key in node.body:
                    if isinstance(key, ast.Assign):
                        for target in key.targets:
                            if isinstance(target, ast.Name):
                                target_name = target.id
                                target_type = ast.unparse(key.value).strip()
                                stub += "%s    %s: %s\n" % (indent, target_name, target_type)
                            elif isinstance(target, ast.Subscript):
                                if isinstance(target.value, ast.Name):
                                    target_name = target.value.id
                                target_type = ast.unparse(key.value).strip()
                                stub += "%s    %s: %s\n" % (indent, target_name, target_type)
                    elif isinstance(key, ast.AnnAssign):
                        target = key.target
                        if isinstance(target, ast.Name):
                            target_name = target.id
                            target_type = ast.unparse(key.annotation).strip()
                            stub += "%s    %s: %s\n" % (indent, target_name, target_type)
                        elif isinstance(target, ast.Subscript):
                            if isinstance(target.value, ast.Name):
                                target_name = target.value.id
                            target_type = ast.unparse(key.annotation).strip()
                            stub += "%s    %s: %s\n" % (indent, target_name, target_type)
            elif case == "Exception":
                if not any((isinstance(n, ast.FunctionDef) for n in node.body)):
                    stub = "%sclass %s(Exception): ..." % (indent, class_name)
                else:
                    stub = "%sclass %s(Exception):" % (indent, class_name)
            elif case == "NamedTuple":
                stub = "%sclass %s(NamedTuple):\n" % (indent, class_name)
                self.imports_output.add("from typing import NamedTuple")
            else:
                is_dataclass = any((isinstance(n, ast.AnnAssign) for n in node.body))
                if is_dataclass:
                    stub = "%s@dataclass\n" % indent
                    self.imports_output.add("from dataclasses import dataclass")
                class_def = "%sclass %s" % (indent, class_name)
                bases = []
                for base in node.bases:
                    if isinstance(base, ast.Name):
                        bases.append(base.id)
                    elif isinstance(base, ast.Subscript):
                        if (
                            isinstance(base.value, ast.Name)
                            and base.value.id == "Generic"
                        ):
                            continue
                        base_name = ast.unparse(base).strip()
                        bases.append(base_name)
                if class_has_generic:
                    bases.append("Generic[%s]" % ', '.join(generic_types))
                if bases:
                    class_def += "(%s)" % ', '.join(bases)
                class_def += ":"
                stub += class_def
            self.stubs.append(stub)
            methods_or_classes = [
                n for n in node.body if isinstance(n, (ast.FunctionDef, ast.ClassDef))
            ]
            if methods_or_classes:
                self.stubs.append("\n")
                class_nodes = []
                method_nodes = []
                for child in methods_or_classes:
                    if isinstance(child, ast.ClassDef):
                        class_nodes.append(child)
                    else:
                        method_nodes.append(child)
                for class_node in class_nodes:
                    self.visit(class_node)
                if class_nodes and method_nodes:
                    self.stubs.append("\n")
                for method_node in method_nodes:
                    self.visit(method_node)
            self.indentation_level = previous_indent
            self.in_class = previous_in_class
            if not previous_in_class:
                self.stubs.append("\n")

        def special_cases(self, node):
            for obj in node.bases:
                ob_instance = isinstance(obj, ast.Name)
                if ob_instance:
                    if obj.id == "TypedDict":
                        return "TypedDict"
                    elif obj.id == "Exception":
                        return "Exception"
                    elif obj.id == "NamedTuple":
                        return "NamedTuple"
                    else:
                        return False
            return False

        def get_arg_type(self, arg_node):
            selfs = ["self", "cls"]
            if arg_node.arg in selfs:
                if (
                    arg_node.arg == "self"
                    and "typing_extensions" not in self.imports_helper_dict
                ):
                    self.imports_helper_dict["typing_extensions"] = set()
                    self.imports_helper_dict["typing_extensions"].add("Self")
                return "Self" if arg_node.arg == "self" else arg_node.arg
            elif arg_node.annotation:
                unparsed = ast.unparse(arg_node.annotation).strip()
                if (
                    isinstance(arg_node.annotation, ast.Name)
                    and arg_node.annotation.id in self.typevars
                ):
                    return arg_node.annotation.id
                if unparsed.startswith("typing."):
                    type_name = unparsed.split(".")[-1]
                    if "typing" not in self.imports_helper_dict:
                        self.imports_helper_dict["typing"] = set()
                    self.imports_helper_dict["typing"].add(type_name)
                    return type_name
                return unparsed
            else:
                if "typing" not in self.imports_helper_dict:
                    self.imports_helper_dict["typing"] = set()
                self.imports_helper_dict["typing"].add("Any")
                return "Any"

        def get_return_type(self, return_node):
            if return_node:
                unparsed = ast.unparse(return_node).strip()
                if (
                    isinstance(return_node, ast.Name)
                    and return_node.id in self.typevars
                ):
                    return return_node.id
                if unparsed.startswith("typing."):
                    type_name = unparsed.split(".")[-1]
                    if "typing" not in self.imports_helper_dict:
                        self.imports_helper_dict["typing"] = set()
                    self.imports_helper_dict["typing"].add(type_name)
                    return type_name
                return unparsed
            else:
                if "typing" not in self.imports_helper_dict:
                    self.imports_helper_dict["typing"] = set()
                self.imports_helper_dict["typing"].add("Any")
                return "Any"

        def visit_AnnAssign(self, node):
            target = node.target
            if isinstance(node.annotation, ast.Name):
                target_type = node.annotation.id
            else:
                target_type = ast.unparse(node.annotation).strip()
                if target_type.startswith("typing."):
                    type_name = target_type.split(".")[-1]
                    if "typing" not in self.imports_helper_dict:
                        self.imports_helper_dict["typing"] = set()
                    self.imports_helper_dict["typing"].add(type_name)
                    target_type = type_name
            if not self.in_class:
                if isinstance(target, ast.Name):
                    target_name = target.id
                    if node.value is not None:
                        value_str = ast.unparse(node.value).strip()
                        stub = "%s: %s = %s\n" % (target_name, target_type, value_str)
                    else:
                        stub = "%s: %s\n" % (target_name, target_type)
                    self.stubs.append(stub)
                    return
            if self.in_class:
                indent = "    " * (self.indentation_level + 1)
                if isinstance(node.annotation, ast.Subscript):
                    if isinstance(target, ast.Name):
                        stub = "%s%s: %s\n" % (indent, target.id, target_type)
                    elif isinstance(target, ast.Subscript):
                        if isinstance(target.value, ast.Name):
                            target_name = target.value.id
                        stub = "%s%s: %s\n" % (indent, target_name, target_type)
                elif isinstance(node.annotation, ast.Name):
                    if isinstance(target, ast.Name):
                        stub = "%s%s: %s\n" % (indent, target.id, target_type)
                    elif isinstance(target, ast.Subscript):
                        if isinstance(target.value, ast.Name):
                            target_name = target.value.id
                        stub = "%s%s: %s\n" % (indent, target_name, target_type)
                elif isinstance(node.annotation, ast.BinOp):
                    if isinstance(target, ast.Name):
                        stub = "%s%s: %s\n" % (indent, target.id, target_type)
                    elif isinstance(target, ast.Subscript):
                        if isinstance(target.value, ast.Name):
                            target_name = target.value.id
                            stub = "%s%s: %s\n" % (indent, target_name, target_type)
                        else:
                            stub = "%s%s: %s\n" % (indent, ast.unparse(target), target_type)
                    else:
                        stub = "%s%s: %s\n" % (indent, ast.unparse(target), target_type)
                else:
                    raise NotImplementedError(
                        "Type %s not implemented, report this issue" % type(node.annotation)
                    )
                self.stubs.append(stub)

        def generate_imports(self):
            imports = []
            if sys.version_info >= (3, 7):
                imports.append("from __future__ import annotations\n")
            sorted_items = sorted(self.imports_helper_dict.items())
            for module, names in sorted_items:
                if names:
                    imports.append("from %s import %s\n" % (module, ', '.join(sorted(names))))
            for imp in sorted(self.imports_output):
                if imp != "from __future__ import annotations":
                    imports.append("%s\n" % imp)
            return "".join(imports) + "\n" if imports else ""

    stub_generator = StubGenerator()
    stub_generator.visit(tree)
    out_str = ""
    imports_str = stub_generator.generate_imports()
    if imports_str:
        out_str += imports_str
    for stub in stub_generator.stubs:
        out_str += stub
    if text_only:
        return out_str
    else:
        with open(output_file_path, "w") as output_file:
            output_file.write(out_str)
        return None


def generate_stub(source_file_path, output_file_path, text_only=False):
    with open(source_file_path, "r", encoding="utf-8") as source_file:
        source_code = source_file.read()
    return generate_stub_from_source(
        source_code=source_code, output_file_path=output_file_path, text_only=text_only
    )


def generate_text_stub(source_file_path):
    stubs = generate_stub(source_file_path, "", text_only=True)
    if stubs:
        return stubs
    else:
        raise ValueError("Stub generation failed")



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Variable type for true/false Variables.

Usage example::

    opts = Variables()
    opts.Add(BoolVariable('embedded', 'build for an embedded system', False))
    env = Environment(variables=opts)
    if env['embedded']:
        ...
"""

from __future__ import annotations

from typing import Callable

import SCons.Errors

__all__ = ['BoolVariable',]

TRUE_STRINGS = ('y', 'yes', 'true', 't', '1', 'on', 'all')
FALSE_STRINGS = ('n', 'no', 'false', 'f', '0', 'off', 'none')


def _text2bool(val):
    """Convert boolean-like string to boolean.

    If *val* looks like it expresses a bool-like value, based on
    the :const:`TRUE_STRINGS` and :const:`FALSE_STRINGS` tuples,
    return the appropriate value.

    This is usable as a converter function for SCons Variables.

    Raises:
        ValueError: if *val* cannot be converted to boolean.
    """
    if isinstance(val, bool):
        # mainly for the subst=False case: default might be a bool
        return val
    lval = val.lower()
    if lval in TRUE_STRINGS:
        return True
    if lval in FALSE_STRINGS:
        return False
    # TODO: leave this check to validator?
    raise ValueError(f"Invalid value for boolean variable: {val!r}")


def _validator(key, val, env):
    """Validate that the value of *key* in *env* is a boolean.

    Parameter *val* is not used in the check.

    Usable as a validator function for SCons Variables.

    Raises:
        KeyError: if *key* is not set in *env*
        UserError: if the value of *key* is not ``True`` or ``False``.

    """
    if env[key] not in (True, False):
        msg = f'Invalid value for boolean variable {key!r}: {env[key]}'
        raise SCons.Errors.UserError(msg) from None

# lint: W0622: Redefining built-in 'help' (redefined-builtin)
def BoolVariable(key, help, default):
    """Return a tuple describing a boolean SCons Variable.

    The input parameters describe a boolean variable, using a string
    value as described by :const:`TRUE_STRINGS` and :const:`FALSE_STRINGS`.
    Returns a tuple including the correct converter and validator.
    The *help* text will have ``(yes|no)`` automatically appended to
    show the valid values. The result is usable as input to
    :meth:`~SCons.Variables.Variables.Add`.
    """
    help = f'{help} (yes|no)'
    return key, help, default, _validator, _text2bool

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Codes for operations.

There are unary and binary operations. Many of them have specializations and
of course types could play into it. Then there is also the added difficulty of
in-place assignments, which have other operation variants.
"""

from nuitka.nodes.shapes.StandardShapes import tshape_unknown

from .BinaryOperationHelperDefinitions import (
    getCodeNameForBinaryOperation,
    getNonSpecializedBinaryOperations,
    getSpecializedBinaryOperations,
)
from .c_types.CTypeBooleans import CTypeBool
from .c_types.CTypeNuitkaBooleans import CTypeNuitkaBoolEnum
from .c_types.CTypeNuitkaVoids import CTypeNuitkaVoidEnum
from .c_types.CTypePyObjectPointers import CTypePyObjectPtr
from .c_types.CTypeVoids import CTypeVoid
from .CodeHelpers import (
    generateChildExpressionsCode,
    generateExpressionCode,
    withObjectCodeTemporaryAssignment,
)
from .CodeHelperSelection import selectCodeHelper
from .ErrorCodes import (
    getErrorExitBoolCode,
    getErrorExitCode,
    getReleaseCodes,
    getTakeReferenceCode,
)
from .ExpressionCTypeSelectionHelpers import decideExpressionCTypes


def generateOperationBinaryCode(to_name, expression, emit, context):
    _getBinaryOperationCode(
        to_name=to_name,
        operator=expression.getOperator(),
        inplace=expression.isInplaceSuspect(),
        needs_check=expression.mayRaiseExceptionOperation(),
        left=expression.subnode_left,
        right=expression.subnode_right,
        source_ref=expression.source_ref,
        emit=emit,
        context=context,
    )


def generateOperationNotCode(to_name, expression, emit, context):
    # TODO: We badly need to support target boolean C type here, or else an object is created from the argument.
    (arg_name,) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    res_name = context.getIntResName()

    emit("%s = CHECK_IF_TRUE(%s);" % (res_name, arg_name))

    getErrorExitBoolCode(
        condition="%s == -1" % res_name,
        release_name=arg_name,
        needs_check=expression.subnode_operand.mayRaiseExceptionBool(BaseException),
        emit=emit,
        context=context,
    )

    to_name.getCType().emitAssignmentCodeFromBoolCondition(
        to_name=to_name, condition="%s == 0" % res_name, emit=emit
    )


def generateOperationUnaryCode(to_name, expression, emit, context):
    (arg_name,) = generateChildExpressionsCode(
        expression=expression, emit=emit, context=context
    )

    _getUnaryOperationCode(
        to_name=to_name,
        expression=expression,
        operator=expression.getOperator(),
        arg_name=arg_name,
        needs_check=expression.mayRaiseExceptionOperation(),
        emit=emit,
        context=context,
    )


def _getBinaryOperationCode(
    to_name, operator, inplace, left, right, needs_check, source_ref, emit, context
):
    # This is detail rich stuff, encoding the complexity of what helpers are
    # available, and can be used as a fallback.
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    (
        _unknown_types,
        needs_argument_swap,
        left_shape,
        right_shape,
        left_c_type,
        right_c_type,
    ) = decideExpressionCTypes(
        left=left,
        right=right,
        may_swap_arguments=(
            "never"
            if inplace
            else (
                "number"
                if operator in ("Add", "Mult", "BitOr", "BitAnd", "BitXor")
                else "never"
            )
        ),
        context=context,
    )

    prefix = "%s_OPERATION_%s" % (
        "INPLACE" if operator[0] == "I" else "BINARY",
        getCodeNameForBinaryOperation(operator),
    )

    specialized_helpers_set = getSpecializedBinaryOperations(operator)
    non_specialized_helpers_set = getNonSpecializedBinaryOperations(operator)

    report_missing = True

    helper_type = target_type = None if operator[0] == "I" else to_name.getCType()

    if helper_type is not None:
        if needs_check and helper_type is not None:
            # If an exception may occur, we do not have the "NVOID" helpers though, we
            # instead can use the CTypeNuitkaBoolEnum that will easily convert to
            # it.

            if helper_type is CTypeNuitkaVoidEnum:
                helper_type = CTypeNuitkaBoolEnum
                report_missing = False
        else:
            # If no exception may occur, we do not have the "VOID" helpers though, we
            # instead can use the CTypeBool that will easily convert to it.

            # If no exception can occur, do not require a helper that can indicate
            # it, but use the one that produces simpler code, this means we can
            # avoid the CTypeNuitkaBoolEnum (NBOOL) helpers except for things that
            # can really raise. Once we have expression for types depending on the
            # value to raise or not, this will get us into trouble, due to using a
            # fallback
            # helper_type = CTypeBool

            # TODO: For now to achieve old behavior, we are going to change to
            # CBOOL for those that cannot raise later.
            if helper_type is CTypeVoid:
                helper_type = CTypeNuitkaBoolEnum
                report_missing = False

    # If a more specific C type was picked that "PyObject *" then we can use that to have the helper.
    helper_type, helper_function = selectCodeHelper(
        prefix=prefix,
        specialized_helpers_set=specialized_helpers_set,
        non_specialized_helpers_set=non_specialized_helpers_set,
        result_type=helper_type,
        left_shape=left_shape,
        right_shape=right_shape,
        left_c_type=left_c_type,
        right_c_type=right_c_type,
        argument_swap=needs_argument_swap,
        report_missing=report_missing,
        source_ref=source_ref,
    )

    # If we failed to find CTypeBool, that should be OK.
    if helper_function is None and target_type is CTypeBool:
        helper_type, helper_function = selectCodeHelper(
            prefix=prefix,
            specialized_helpers_set=specialized_helpers_set,
            non_specialized_helpers_set=non_specialized_helpers_set,
            result_type=CTypeNuitkaBoolEnum,
            left_shape=left_shape,
            right_shape=right_shape,
            left_c_type=left_c_type,
            right_c_type=right_c_type,
            argument_swap=needs_argument_swap,
            report_missing=True,
            source_ref=source_ref,
        )

    if helper_function is None:
        # Give up and warn about it.
        left_c_type = CTypePyObjectPtr
        right_c_type = CTypePyObjectPtr

        helper_type, helper_function = selectCodeHelper(
            prefix=prefix,
            specialized_helpers_set=specialized_helpers_set,
            non_specialized_helpers_set=non_specialized_helpers_set,
            result_type=CTypePyObjectPtr if helper_type is not None else None,
            left_shape=tshape_unknown,
            right_shape=tshape_unknown,
            left_c_type=left_c_type,
            right_c_type=right_c_type,
            argument_swap=False,
            report_missing=True,
            source_ref=source_ref,
        )

        assert helper_function is not None, (left, right)

    left_name = context.allocateTempName(
        "%s_expr_left" % operator.lower(), type_name=left_c_type.c_type
    )
    right_name = context.allocateTempName(
        "%s_expr_right" % operator.lower(), type_name=right_c_type.c_type
    )

    generateExpressionCode(
        to_name=left_name, expression=left, emit=emit, context=context
    )
    generateExpressionCode(
        to_name=right_name, expression=right, emit=emit, context=context
    )

    # We must assume to write to a variable if "inplace" is active, not e.g.
    # a constant reference. That was asserted before calling us.
    if inplace or "INPLACE" in helper_function:
        assert not needs_argument_swap

        res_name = context.getBoolResName()

        # For module variable C type to reference later.
        if left.isExpressionVariableRef() and left.getVariable().isModuleVariable():
            emit("%s = %s;" % (context.getInplaceLeftName(), left_name))

        if not left.isExpressionVariableRefOrTempVariableRef():
            if not context.needsCleanup(left_name):
                getTakeReferenceCode(left_name, emit)

        emit("%s = %s(&%s, %s);" % (res_name, helper_function, left_name, right_name))

        getErrorExitBoolCode(
            condition="%s == false" % res_name,
            release_names=(left_name, right_name),
            needs_check=needs_check,
            emit=emit,
            context=context,
        )

        emit("%s = %s;" % (to_name, left_name))

        if not left.isExpressionVariableRefOrTempVariableRef():
            context.addCleanupTempName(to_name)
    else:
        if needs_argument_swap:
            arg1_name = right_name
            arg2_name = left_name
            arg1_c_type = right_c_type
            arg2_c_type = left_c_type
        else:
            arg1_name = left_name
            arg2_name = right_name
            arg1_c_type = left_c_type
            arg2_c_type = right_c_type

        # May need to convert return value.
        if helper_type is not target_type:
            value_name = context.allocateTempName(
                to_name.code_name + "_" + helper_type.helper_code.lower(),
                type_name=helper_type.c_type,
                unique=to_name.code_name == "tmp_unused",
            )
        else:
            value_name = to_name

        if helper_type.isDualType():
            res_name = context.getBoolResName()

            # TODO: If possible, pass variable storage directly to avoid useless
            # copy.
            emit(
                "%s = %s(&%s, %s%s, %s%s);"
                % (
                    res_name,
                    helper_function,
                    value_name,
                    "&" if arg1_c_type.isDualType() else "",
                    arg1_name,
                    "&" if arg2_c_type.isDualType() else "",
                    arg2_name,
                )
            )

            getErrorExitBoolCode(
                condition="%s == false" % res_name,
                release_names=(left_name, right_name),
                needs_check=needs_check,
                emit=emit,
                context=context,
            )
        else:
            emit(
                "%s = %s(%s, %s);"
                % (
                    value_name,
                    helper_function,
                    arg1_name,
                    arg2_name,
                )
            )

            if value_name.getCType().hasErrorIndicator():
                getErrorExitCode(
                    check_name=value_name,
                    release_names=(left_name, right_name),
                    needs_check=needs_check,
                    emit=emit,
                    context=context,
                )
            else:
                # Otherwise we picked the wrong kind of helper.
                assert not needs_check, value_name.getCType()

                getReleaseCodes(
                    release_names=(left_name, right_name), emit=emit, context=context
                )

        # TODO: Depending on operation, we could not produce a reference, if result *must*
        # be boolean, but then we would have some helpers that do it, and some that do not
        # do it.
        if helper_type.hasReleaseCode():
            context.addCleanupTempName(value_name)

        if value_name is not to_name:
            target_type.emitAssignConversionCode(
                to_name=to_name,
                value_name=value_name,
                # TODO: Right now we don't do conversions here that could fail.
                needs_check=False,
                emit=emit,
                context=context,
            )


unary_operator_codes = {
    "UAdd": ("PyNumber_Positive", 1),
    "USub": ("PyNumber_Negative", 1),
    "Invert": ("PyNumber_Invert", 1),
    "Repr": ("PyObject_Repr", 1),
    "Not": ("UNARY_NOT", 0),
}


def _getUnaryOperationCode(
    to_name, expression, operator, arg_name, needs_check, emit, context
):
    impl_helper, ref_count = unary_operator_codes[operator]

    helper = "UNARY_OPERATION"
    prefix_args = (impl_helper,)

    with withObjectCodeTemporaryAssignment(
        to_name, "op_%s_res" % operator.lower(), expression, emit, context
    ) as value_name:
        emit(
            "%s = %s(%s);"
            % (
                value_name,
                helper,
                ", ".join(str(arg_name) for arg_name in prefix_args + (arg_name,)),
            )
        )

        getErrorExitCode(
            check_name=value_name,
            release_name=arg_name,
            needs_check=needs_check,
            emit=emit,
            context=context,
        )

        if ref_count:
            context.addCleanupTempName(value_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# Plugin dill-compat has this as post load code for the "dill" package.
# Not supposed to be good code,
# pylint: disable=invalid-name,missing-module-docstring,protected-access
# pylint: disable=too-many-branches,too-many-statements,used-before-assignment


try:
    import dill._dill
except ImportError:
    pass
else:
    from types import CodeType

    # spell-checker: ignore kwdefaults,globalvars
    # Nuitka will optimize this away, but VS code will warn about them otherwise.
    from typing import TYPE_CHECKING

    if str is bytes:
        import __builtin__ as builtins  # Python2 code, pylint: disable=I0021,import-error
    else:
        import builtins

    if TYPE_CHECKING:
        compiled_function_tables = {}

    class ForCompiledTypeLookups:
        def for_compiled_type(self):
            pass

    compiled_function = type(ForCompiledTypeLookups.for_compiled_type)
    compiled_method = type(ForCompiledTypeLookups().for_compiled_type)
    assert "__compiled__" in globals()

    dill_version = tuple(int(d) for d in dill.__version__.split("."))

    # Compiled methods need to be created.
    @dill.register(compiled_method)
    def save_compiled_method(pickler, obj):
        if str is not bytes:
            pickler.save_reduce(compiled_method, (obj.__func__, obj.__self__), obj=obj)
        else:
            pickler.save_reduce(
                compiled_method, (obj.im_func, obj.im_self, obj.im_class), obj=obj
            )

    def _create_compiled_function(module_name, func_values):
        if module_name not in compiled_function_tables:
            __import__(module_name)

        # This gets the "_create_compiled_function" of the module and calls it.
        return compiled_function_tables[module_name][1](*func_values)

    # Compiled methods might have to be created as well. This is very closely following
    # the code of "dill._dill.save_function" that is intended to do things for the
    # uncompiled function.
    @dill.register(compiled_function)
    def save_compiled_function(pickler, obj):
        # Complex beast, pylint: disable=too-many-locals

        if dill_version >= (0, 3):
            if not dill._dill._locate_function(obj, pickler):
                if type(obj.__code__) is not CodeType:
                    # Some PyPy builtin functions have no module name, and thus are not
                    # able to be located
                    module_name = getattr(obj, "__module__", None)
                    if module_name is None:
                        module_name = builtins.__name__

                    _module = dill._dill._import_module(module_name, safe=True)

                _recurse = getattr(pickler, "_recurse", None)
                _postproc = getattr(pickler, "_postproc", None)
                _main_modified = getattr(pickler, "_main_modified", None)
                _original_main = getattr(pickler, "_original_main", builtins)

                postproc_list = []
                if _recurse:
                    # recurse to get all globals referred to by obj
                    from dill.detect import globalvars

                    globs_copy = globalvars(obj, recurse=True, builtin=True)

                    # Add the name of the module to the globs dictionary to prevent
                    # the duplication of the dictionary. Pickle the unpopulated
                    # globals dictionary and set the remaining items after the function
                    # is created to correctly handle recursion.
                    globs = {"__name__": obj.__module__}
                else:
                    globs_copy = obj.__globals__

                    # If the globals is the __dict__ from the module being saved as a
                    # session, substitute it by the dictionary being actually saved.
                    if _main_modified and globs_copy is _original_main.__dict__:
                        globs_copy = getattr(pickler, "_main", _original_main).__dict__
                        globs = globs_copy
                    # If the globals is a module __dict__, do not save it in the pickle.
                    elif (
                        globs_copy is not None
                        and obj.__module__ is not None
                        and getattr(
                            dill._dill._import_module(obj.__module__, True),
                            "__dict__",
                            None,
                        )
                        is globs_copy
                    ):
                        globs = globs_copy
                    else:
                        globs = {"__name__": obj.__module__}

                if globs_copy is not None and globs is not globs_copy:
                    # In the case that the globals are copied, we need to ensure that
                    # the globals dictionary is updated when all objects in the
                    # dictionary are already created.
                    glob_ids = {id(g) for g in globs_copy.values()}
                    for stack_element in _postproc:
                        if stack_element in glob_ids:
                            _postproc[stack_element].append(
                                (dill._dill._setitems, (globs, globs_copy))
                            )
                            break
                    else:
                        postproc_list.append(
                            (dill._dill._setitems, (globs, globs_copy))
                        )

                state_dict = {}
                for fattrname in ("__doc__", "__kwdefaults__", "__annotations__"):
                    fattr = getattr(obj, fattrname, None)
                    if fattr is not None:
                        state_dict[fattrname] = fattr
                if obj.__qualname__ != obj.__name__:
                    state_dict["__qualname__"] = obj.__qualname__
                if "__name__" not in globs or obj.__module__ != globs["__name__"]:
                    state_dict["__module__"] = obj.__module__

                pickler.save_reduce(
                    _create_compiled_function,
                    (
                        obj.__module__,
                        # This gets the "_reduce_compiled_function" of the module and calls it.
                        compiled_function_tables[obj.__module__][0](obj),
                    ),
                )
            else:
                name = getattr(obj, "__qualname__", getattr(obj, "__name__", None))
                dill._dill.StockPickler.save_global(pickler, obj, name=name)
        else:
            if not dill._dill._locate_function(obj):
                stack = dill._dill.stack

                _byref = getattr(pickler, "_byref", None)
                _recurse = getattr(pickler, "_recurse", None)
                _memo = (id(obj) in stack) and (_recurse is not None)
                stack[id(obj)] = len(stack), obj

                if str is not bytes:
                    # Python3
                    _super = ("super" in getattr(obj.__code__, "co_names", ())) and (
                        _byref is not None
                    )
                    if _super:
                        pickler._byref = True
                    if _memo:
                        pickler._recurse = False

                    pickler.save_reduce(
                        _create_compiled_function,
                        (
                            obj.__module__,
                            # This gets the "_reduce_compiled_function" of the module and calls it.
                            compiled_function_tables[obj.__module__][0](obj),
                        ),
                    )
                else:
                    # Python2
                    _super = (
                        ("super" in getattr(obj.__code__, "co_names", ()))
                        and (_byref is not None)
                        and getattr(pickler, "_recurse", False)
                    )
                    if _super:
                        pickler._byref = True
                    if _memo:
                        pickler._recurse = False

                    pickler.save_reduce(
                        _create_compiled_function,
                        (
                            obj.__module__,
                            # This gets the "_reduce_compiled_function" of the module and calls it.
                            compiled_function_tables[obj.__module__][0](obj),
                        ),
                    )

                if _super:
                    pickler._byref = _byref
                if _memo:
                    pickler._recurse = _recurse
            else:
                dill._dill.StockPickler.save_global(pickler, obj)

    if __compiled__.standalone:  # pylint: disable=undefined-variable
        builtins.compiled_method = compiled_method
        builtins.compiled_function = compiled_function

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Nodes for async generator objects and their creations.

Async generator are turned into normal functions that create generator objects,
whose implementation lives here. The creation itself also lives here.

"""

from .ChildrenHavingMixins import ChildHavingAsyncgenRefMixin
from .ExpressionBases import ExpressionBase, ExpressionNoSideEffectsMixin
from .FunctionNodes import ExpressionFunctionEntryPointBase


class ExpressionMakeAsyncgenObject(
    ExpressionNoSideEffectsMixin, ChildHavingAsyncgenRefMixin, ExpressionBase
):
    kind = "EXPRESSION_MAKE_ASYNCGEN_OBJECT"

    named_children = ("asyncgen_ref",)

    __slots__ = ("variable_closure_traces",)

    def __init__(self, asyncgen_ref, source_ref):
        assert asyncgen_ref.getFunctionBody().isExpressionAsyncgenObjectBody()

        ChildHavingAsyncgenRefMixin.__init__(self, asyncgen_ref=asyncgen_ref)

        ExpressionBase.__init__(self, source_ref)

        self.variable_closure_traces = []

    def getDetailsForDisplay(self):
        return {"asyncgen": self.subnode_asyncgen_ref.getFunctionBody().getCodeName()}

    def computeExpression(self, trace_collection):
        self.variable_closure_traces = []

        for (
            closure_variable
        ) in self.subnode_asyncgen_ref.getFunctionBody().getClosureVariables():
            trace = trace_collection.getVariableCurrentTrace(closure_variable)
            trace.addNameUsage()

            self.variable_closure_traces.append((closure_variable, trace))

        # TODO: Asyncgen body may know something too.
        return self, None, None

    def getClosureVariableVersions(self):
        return self.variable_closure_traces


class ExpressionAsyncgenObjectBody(ExpressionFunctionEntryPointBase):
    kind = "EXPRESSION_ASYNCGEN_OBJECT_BODY"

    __slots__ = ("qualname_setup", "needs_generator_return_exit")

    def __init__(self, provider, name, code_object, flags, auto_release, source_ref):
        ExpressionFunctionEntryPointBase.__init__(
            self,
            provider=provider,
            name=name,
            code_object=code_object,
            code_prefix="asyncgen",
            flags=flags,
            auto_release=auto_release,
            source_ref=source_ref,
        )

        self.needs_generator_return_exit = False

        self.qualname_setup = None

    def getFunctionName(self):
        return self.name

    def markAsNeedsGeneratorReturnHandling(self):
        self.needs_generator_return_exit = True

    def needsGeneratorReturnExit(self):
        return self.needs_generator_return_exit

    @staticmethod
    def needsCreation():
        return False

    @staticmethod
    def isUnoptimized():
        return False


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
Namespace packages of Python3

"""

import os

from nuitka.nodes.AttributeNodes import makeExpressionAttributeLookup
from nuitka.nodes.CallNodes import ExpressionCallNoKeywords
from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
from nuitka.nodes.ContainerMakingNodes import (
    makeExpressionMakeList,
    makeExpressionMakeTuple,
    makeExpressionMakeTupleOrConstant,
)
from nuitka.nodes.FutureSpecs import FutureSpec
from nuitka.nodes.ImportNodes import (
    ExpressionImportName,
    makeExpressionImportModuleFixed,
    makeExpressionImportModuleNameHard,
)
from nuitka.nodes.ModuleAttributeNodes import ExpressionModuleAttributeFileRef
from nuitka.nodes.ModuleNodes import CompiledPythonNamespacePackage
from nuitka.nodes.VariableNameNodes import StatementAssignmentVariableName
from nuitka.options.Options import getFileReferenceMode
from nuitka.PythonVersions import python_version

from .FutureSpecState import popFutureSpec, pushFutureSpec
from .TreeHelpers import (
    buildNode,
    makeStatementsSequenceFromStatement,
    parseSourceCodeToAst,
)
from .VariableClosure import completeVariableClosures


def _makeCall(module_name, import_name, attribute_name, source_ref, *args):
    return ExpressionCallNoKeywords(
        called=makeExpressionAttributeLookup(
            expression=makeExpressionImportModuleNameHard(
                module_name=module_name,
                import_name=import_name,
                module_guaranteed=True,
                source_ref=source_ref,
            ),
            attribute_name=attribute_name,
            source_ref=source_ref,
        ),
        args=makeExpressionMakeTupleOrConstant(
            elements=args, user_provided=True, source_ref=source_ref
        ),
        source_ref=source_ref,
    )


def _getNameSpacePathExpression(package, source_ref):
    """Create the __path__ expression for a package."""

    reference_mode = getFileReferenceMode()

    if reference_mode == "original":
        return makeConstantRefNode(
            constant=[package.getCompileTimeDirectory()],
            source_ref=source_ref,
        )
    elif reference_mode == "frozen":
        return makeConstantRefNode(
            constant=[],
            source_ref=source_ref,
        )
    else:
        file_ref_node = ExpressionModuleAttributeFileRef(
            variable=package.getVariableForReference("__file__"),
            source_ref=source_ref,
        )

        if package.isCompiledPythonNamespacePackage():
            elements = [file_ref_node]
        else:
            elements = [
                ExpressionCallNoKeywords(
                    called=makeExpressionAttributeLookup(
                        expression=makeExpressionImportModuleNameHard(
                            module_name="os",
                            import_name="path",
                            module_guaranteed=True,
                            source_ref=source_ref,
                        ),
                        attribute_name="dirname",
                        source_ref=source_ref,
                    ),
                    args=makeExpressionMakeTuple(
                        elements=(file_ref_node,),
                        source_ref=source_ref,
                    ),
                    source_ref=source_ref,
                )
            ]

        if package.canHaveExternalImports():
            parts = package.getFullName().asString().split(".")

            for count in range(len(parts)):
                path_part = _makeCall(
                    "os",
                    "environ",
                    "get",
                    source_ref,
                    makeConstantRefNode(
                        constant="NUITKA_PACKAGE_%s" % "_".join(parts[: count + 1]),
                        source_ref=source_ref,
                    ),
                    makeConstantRefNode(
                        constant=os.path.sep + "not_existing", source_ref=source_ref
                    ),
                )

                if parts[count + 1 :]:
                    path_part = _makeCall(
                        "os",
                        "path",
                        "join",
                        source_ref,
                        path_part,
                        makeConstantRefNode(
                            constant=os.path.join(*parts[count + 1 :]),
                            source_ref=source_ref,
                        ),
                    )

                elements.append(path_part)

        return makeExpressionMakeList(elements=tuple(elements), source_ref=source_ref)


def createPathAssignment(package, source_ref):
    return StatementAssignmentVariableName(
        provider=package,
        variable_name="__path__",
        source=_getNameSpacePathExpression(package=package, source_ref=source_ref),
        source_ref=source_ref,
    )


def _getPathFinderFunction(package, source_ref):
    # TODO: This could be a shared helper function, but we currently cannot do these.
    ast_tree = parseSourceCodeToAst(
        source_code="_path_finder = lambda *args, **kw: None",
        module_name=package.getFullName(),
        filename=source_ref.getFilename(),
        line_offset=0,
    )

    pushFutureSpec(package.getFullName())
    statement = buildNode(package, ast_tree.body[0], source_ref)
    popFutureSpec()

    return statement.subnode_source


def createPython3NamespacePath(package, source_ref):

    module_name = (
        "_frozen_importlib" if python_version < 0x350 else "_frozen_importlib_external"
    )

    return StatementAssignmentVariableName(
        provider=package,
        variable_name="__path__",
        source=ExpressionCallNoKeywords(
            called=ExpressionImportName(
                module=makeExpressionImportModuleFixed(
                    using_module_name=package.getFullName(),
                    module_name=module_name,
                    value_name=module_name,
                    source_ref=source_ref,
                ),
                import_name="_NamespacePath",
                level=0,
                source_ref=source_ref,
            ),
            args=makeExpressionMakeTupleOrConstant(
                elements=(
                    makeConstantRefNode(
                        constant=package.getFullName().asString(),
                        user_provided=True,
                        source_ref=source_ref,
                    ),
                    _getNameSpacePathExpression(package=package, source_ref=source_ref),
                    _getPathFinderFunction(package=package, source_ref=source_ref),
                ),
                user_provided=True,
                source_ref=source_ref,
            ),
            source_ref=source_ref,
        ),
        source_ref=source_ref,
    )


def createNamespacePackage(module_name, reason, is_top, source_ref):
    package = CompiledPythonNamespacePackage(
        module_name=module_name,
        reason=reason,
        is_top=is_top,
        mode="compiled",
        future_spec=FutureSpec(use_annotations=False),
        source_ref=source_ref,
    )

    if python_version >= 0x300:
        statement = createPython3NamespacePath(package=package, source_ref=source_ref)
    else:
        statement = createPathAssignment(package, source_ref)

    package.setChildBody(makeStatementsSequenceFromStatement(statement=statement))

    completeVariableClosures(package)

    return package


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Search modes for Nuitka's test runner.

The test runner can handle found errors, skip tests, etc. with search
modes, which are implemented here.
"""

import os
import sys
from fnmatch import fnmatch

from nuitka.__past__ import md5
from nuitka.utils.FileOperations import (
    areSamePaths,
    getFileContents,
    putTextFileContents,
)


class SearchMode(object):
    # Lots of details, to keep track of pylint: disable=too-many-instance-attributes

    __slots__ = (
        "active",
        "start_at",
        "only",
        "skip",
        "coverage",
        "abort_on_error",
        "start_dir",
        "logger",
        "may_fail",
        "verifications",
        "failed",
        "had_match",
        "cache_filename",
        "resume_from",
        "max_failures",
    )

    def __init__(
        self,
        logger,
        start_at,
        start_dir,
        resume=False,
        only=False,
        skip=False,
        abort_on_error=True,
        coverage=False,
        max_failures=None,
    ):
        self.active = False
        self.start_at = start_at
        self.only = only
        self.skip = skip
        self.coverage = coverage
        self.had_match = False
        self.abort_on_error = abort_on_error
        self.max_failures = max_failures

        self.start_dir = start_dir
        self.logger = logger
        self.may_fail = []

        if only and not abort_on_error:
            self.logger.sysexit("Error, cannot combine --only-one and --all.")

        self.verifications = 0
        self.failed = []

        # We are going to produce a hash from the command line script we are running,
        # so we have a unique place to store the resume state.
        tests_path = os.path.normcase(os.path.abspath(sys.modules["__main__"].__file__))
        version = sys.version

        if str is not bytes:
            tests_path = tests_path.encode("utf8")
            version = version.encode("utf8")

        case_hash = md5(tests_path)
        case_hash.update(version)

        from .Common import getTestingCacheDir

        cache_filename = os.path.join(getTestingCacheDir(), case_hash.hexdigest())

        self.cache_filename = cache_filename

        if resume and os.path.exists(cache_filename):
            self.resume_from = getFileContents(cache_filename) or None
        else:
            self.resume_from = None

    def consider(self, dirname, filename):
        if self.active and self.only:
            return False

        # Check if we become active
        if not self.active:
            if self.only and self.had_match:
                return False

            if self.start_at is None and not self.resume_from:
                self.active = True
            elif self.resume_from:
                if self._match(dirname, filename, self.resume_from):
                    self.active = True
                    if self.skip:
                        return False
            elif self.start_at:
                if self._match(dirname, filename, self.start_at):
                    self.active = True

        if self.active:
            # If active, we still filter by pattern if one was given.
            if self.start_at and not self._match(dirname, filename, self.start_at):
                return False

            # If we are active, we can save the current file as the one to resume
            # from.
            self._saveResume(dirname, filename)

            if self.only:
                self.active = False

            self.had_match = True
            self.verifications += 1
            return True

        return False

    def _saveResume(self, dirname, filename):
        parts = [dirname, filename]

        while None in parts:
            parts.remove(None)
        assert parts

        path = os.path.join(*parts)

        putTextFileContents(self.cache_filename, contents=path)

    def abortOnFinding(self, dirname, filename):
        if not self._abortOnFinding(dirname, filename):
            return False

        self.failed.append(filename)

        if self.abort_on_error:
            return True

        if self.max_failures is not None and len(self.failed) >= self.max_failures:
            self.logger.warning("Max failures reached.")
            return True

        return False

    def _abortOnFinding(self, dirname, filename):
        for candidate in self.may_fail:
            if self._match(dirname, filename, candidate):
                return False

        return True

    def onErrorDetected(self, message):
        self.finish(success=False)
        return self.exit(message)

    def finish(self, success=True):
        if not self.active and not self.had_match:
            return self.exit("Error, became never active.")

        if success and os.path.exists(self.cache_filename):
            os.unlink(self.cache_filename)

        print(
            "Ran %d test cases (%d failures)." % (self.verifications, len(self.failed))
        )

        if self.failed:
            return self.exit("Error, tests %s failed." % str(self.failed))

    def getExtraFlags(self, dirname, filename):
        # pylint: disable=unused-argument

        if self.coverage:
            return ["coverage"]
        return []

    def isCoverage(self):
        return self.coverage

    def mayFailFor(self, *names):
        self.may_fail += names

    def _match(self, dirname, filename, candidate):
        parts = [dirname, filename]

        while None in parts:
            parts.remove(None)
        assert parts

        path = os.path.join(*parts)

        candidates = (
            dirname,
            filename,
            filename.rsplit(".", 1)[0],
            filename.rsplit(".", 1)[0].replace("Test", ""),
            path,
            path.rsplit(".", 1)[0],
            path.rsplit(".", 1)[0].replace("Test", ""),
        )

        if candidate.rstrip("/") in candidates:
            return True

        if areSamePaths(os.path.join(self.start_dir, candidate), filename):
            return True

        for c in candidates:
            if c is None:
                continue

            if fnmatch(c, candidate):
                return True

        return False

    def exit(self, message):
        return self.logger.sysexit(message)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Helper functions for parsing the AST nodes and building the Nuitka node tree."""

import __future__

import ast

from nuitka.Constants import createConstantDict
from nuitka.Errors import CodeTooComplexCode
from nuitka.nodes.CallNodes import makeExpressionCall
from nuitka.nodes.CodeObjectSpecs import CodeObjectSpec
from nuitka.nodes.ConstantRefNodes import makeConstantRefNode
from nuitka.nodes.ContainerMakingNodes import makeExpressionMakeTupleOrConstant
from nuitka.nodes.DictionaryNodes import makeExpressionMakeDict
from nuitka.nodes.ExceptionNodes import StatementReraiseException
from nuitka.nodes.FrameNodes import (
    StatementsFrameAsyncgen,
    StatementsFrameClass,
    StatementsFrameCoroutine,
    StatementsFrameFunction,
    StatementsFrameGenerator,
    StatementsFrameModule,
)
from nuitka.nodes.KeyValuePairNodes import (
    makeKeyValuePairExpressionsFromKwArgs,
)
from nuitka.nodes.NodeBases import NodeBase
from nuitka.nodes.NodeMakingHelpers import mergeStatements
from nuitka.nodes.StatementNodes import StatementsSequence
from nuitka.options.Options import hasPythonFlagNoDocStrings
from nuitka.PythonVersions import python_version
from nuitka.States import states
from nuitka.Tracing import optimization_logger, printLine


def dump(node):
    printLine(ast.dump(node))


def getKind(node):
    return node.__class__.__name__.rsplit(".", 1)[-1]


def extractDocFromBody(node):
    body = node.body
    doc = None

    # Work around "ast.get_docstring" breakage.
    if body and getKind(body[0]) == "Expr":
        if getKind(body[0].value) == "Str":  # python3.7 or earlier
            doc = body[0].value.s
            body = body[1:]
        elif getKind(body[0].value) == "Constant":  # python3.8
            # Only strings should be used, but all other constants can immediately be ignored,
            # it seems that e.g. Ellipsis is common.
            if type(body[0].value.value) is str:
                doc = body[0].value.value
            body = body[1:]

        if hasPythonFlagNoDocStrings():
            doc = None

    if doc is not None and python_version >= 0x3D0:
        doc = doc.lstrip()

    return body, doc


def parseSourceCodeToAst(source_code, module_name, filename, line_offset):
    # Workaround: ast.parse cannot cope with some situations where a file is not
    # terminated by a new line.
    if not source_code.endswith("\n"):
        source_code = source_code + "\n"

    try:
        body = ast.parse(source_code, filename)
    except RuntimeError as e:
        if "maximum recursion depth" in e.args[0]:
            raise CodeTooComplexCode(module_name, filename)

        raise

    assert getKind(body) == "Module"

    if line_offset > 0:
        ast.increment_lineno(body, line_offset)

    return body


def detectFunctionBodyKind(nodes, start_value=None):
    # This is a complex mess, following the scope means a lot of checks need
    # to be done. pylint: disable=too-many-branches,too-many-statements

    indications = set()
    if start_value is not None:
        indications.add(start_value)

    flags = set()

    def _checkCoroutine(field):
        """Check only for co-routine nature of the field and only update that."""
        # TODO: This is clumsy code, trying to achieve what non-local does for
        # Python2 as well.

        old = set(indications)
        indications.clear()

        _check(field)

        if "Coroutine" in indications:
            old.add("Coroutine")

        indications.clear()
        indications.update(old)

    def _check(node):
        node_class = node.__class__

        if node_class is ast.Yield:
            indications.add("Generator")
        elif python_version >= 0x300 and node_class is ast.YieldFrom:
            indications.add("Generator")
        elif python_version >= 0x350 and node_class in (ast.Await, ast.AsyncWith):
            indications.add("Coroutine")

        # Recurse to children, but do not cross scope boundary doing so.
        if node_class is ast.ClassDef:
            for name, field in ast.iter_fields(node):
                if name in ("name", "body"):
                    pass
                elif name in ("bases", "decorator_list", "keywords", "type_params"):
                    for child in field:
                        _check(child)
                elif name == "starargs":  # spell-checker: ignore starargs
                    if field is not None:
                        _check(field)
                elif name == "kwargs":
                    if field is not None:
                        _check(field)
                else:
                    assert False, (name, field, ast.dump(node))
        elif node_class in (ast.FunctionDef, ast.Lambda) or (
            python_version >= 0x350 and node_class is ast.AsyncFunctionDef
        ):
            for name, field in ast.iter_fields(node):
                if name in ("name", "body"):
                    pass
                elif name in ("bases", "decorator_list", "type_params"):
                    for child in field:
                        _check(child)
                elif name == "args":
                    for child in field.defaults:
                        _check(child)

                    if python_version >= 0x300:
                        for child in node.args.kw_defaults:
                            if child is not None:
                                _check(child)

                        for child in node.args.args:
                            if child.annotation is not None:
                                _check(child.annotation)

                elif name == "returns":
                    if field is not None:
                        _check(field)
                elif name == "type_comment":
                    # Python3.8: We don't have structure here.
                    assert field is None or type(field) is str
                else:
                    assert False, (name, field, ast.dump(node))
        elif node_class is ast.GeneratorExp:
            for name, field in ast.iter_fields(node):
                if name == "name":
                    pass
                elif name in ("body", "comparators", "elt"):
                    if python_version >= 0x370:
                        _checkCoroutine(field)
                elif name == "generators":
                    _check(field[0].iter)

                    # New syntax in 3.7 allows these to be present in functions not
                    # declared with "async def", so we need to check them, but
                    # only if top level.
                    if python_version >= 0x370 and node in nodes:
                        for gen in field:
                            if gen.is_async:
                                indications.add("Coroutine")
                                break

                            if _checkCoroutine(gen):
                                break
                else:
                    assert False, (name, field, ast.dump(node))
        elif node_class is ast.ListComp and python_version >= 0x300:
            for name, field in ast.iter_fields(node):
                if name in ("name", "body", "comparators"):
                    pass
                elif name == "generators":
                    if python_version < 0x3B0:
                        _check(field[0].iter)
                    else:
                        _check(field[0])
                elif name in ("body", "elt"):
                    _check(field)
                else:
                    assert False, (name, field, ast.dump(node))
        elif python_version >= 0x270 and node_class is ast.SetComp:
            for name, field in ast.iter_fields(node):
                if name in ("name", "body", "comparators", "elt"):
                    pass
                elif name == "generators":
                    _check(field[0].iter)
                else:
                    assert False, (name, field, ast.dump(node))
        elif python_version >= 0x270 and node_class is ast.DictComp:
            for name, field in ast.iter_fields(node):
                if name in ("name", "body", "comparators", "key", "value"):
                    pass
                elif name == "generators":
                    _check(field[0].iter)
                else:
                    assert False, (name, field, ast.dump(node))
        elif python_version >= 0x370 and node_class is ast.comprehension:
            for name, field in ast.iter_fields(node):
                if name in ("name", "target"):
                    pass
                elif name == "iter":
                    # Top level comprehension iterators do not influence those.
                    if node not in nodes:
                        _check(field)
                elif name == "ifs":
                    for child in field:
                        _check(child)
                elif name == "is_async":
                    if field:
                        indications.add("Coroutine")
                else:
                    assert False, (name, field, ast.dump(node))
        elif node_class is ast.Name:
            if python_version >= 0x300 and node.id == "super":
                flags.add("has_super")
        elif python_version < 0x300 and node_class is ast.Exec:
            flags.add("has_exec")

            if node.globals is None:
                flags.add("has_unqualified_exec")

            for child in ast.iter_child_nodes(node):
                _check(child)
        elif python_version < 0x300 and node_class is ast.ImportFrom:
            for import_desc in node.names:
                if import_desc.name[0] == "*":
                    flags.add("has_exec")
            for child in ast.iter_child_nodes(node):
                _check(child)
        else:
            for child in ast.iter_child_nodes(node):
                _check(child)

    for node in nodes:
        _check(node)

    if indications:
        if "Coroutine" in indications and "Generator" in indications:
            function_kind = "Asyncgen"
        else:
            # If we found something, make sure we agree on all clues.
            assert len(indications) == 1, indications
            function_kind = indications.pop()
    else:
        function_kind = "Function"

    return function_kind, flags


build_nodes_args3 = None
build_nodes_args2 = None
build_nodes_args1 = None


def setBuildingDispatchers(path_args3, path_args2, path_args1):
    # Using global here, as this is really a singleton, in the form of a module,
    # and this is to break the cyclic dependency it has, pylint: disable=global-statement

    global build_nodes_args3, build_nodes_args2, build_nodes_args1

    build_nodes_args3 = path_args3
    build_nodes_args2 = path_args2
    build_nodes_args1 = path_args1


def buildNode(provider, node, source_ref, allow_none=False):
    # too many exception handlers, pylint: disable=too-many-branches

    if node is None and allow_none:
        return None

    try:
        kind = getKind(node)

        if hasattr(node, "lineno"):
            source_ref = source_ref.atLineNumber(node.lineno)

        if kind in build_nodes_args3:
            result = build_nodes_args3[kind](
                provider=provider, node=node, source_ref=source_ref
            )
        elif kind in build_nodes_args2:
            result = build_nodes_args2[kind](node=node, source_ref=source_ref)
        elif kind in build_nodes_args1:
            result = build_nodes_args1[kind](source_ref=source_ref)
        elif kind == "Pass":
            result = None
        else:
            assert False, ast.dump(node)

        if result is None and allow_none:
            return None

        assert isinstance(result, NodeBase), result

        return result
    except SyntaxError:
        raise
    except RuntimeError:
        # Very likely the stack overflow, which we will turn into too complex
        # code exception, don't warn about it with a code dump then.
        raise
    except KeyboardInterrupt:
        # User interrupting is not a problem with the source, but tell where
        # we got interrupted.
        optimization_logger.info("Interrupted at '%s'." % source_ref)
        raise
    except SystemExit:
        optimization_logger.warning("Problem at '%s'." % source_ref.getAsString())
        raise
    except:
        optimization_logger.warning(
            "Problem at '%s' with %s." % (source_ref.getAsString(), ast.dump(node))
        )
        raise


def buildNodeList(provider, nodes, source_ref, allow_none=False):
    if nodes is not None:
        result = []

        for node in nodes:
            if hasattr(node, "lineno"):
                node_source_ref = source_ref.atLineNumber(node.lineno)
            else:
                node_source_ref = source_ref

            entry = buildNode(provider, node, node_source_ref, allow_none)

            if entry is not None:
                result.append(entry)

        return result
    else:
        return []


def buildNodeTuple(provider, nodes, source_ref, allow_none=False):
    if nodes is not None:
        result = []

        for node in nodes:
            if hasattr(node, "lineno"):
                node_source_ref = source_ref.atLineNumber(node.lineno)
            else:
                node_source_ref = source_ref

            entry = buildNode(provider, node, node_source_ref, allow_none)

            if entry is not None:
                result.append(entry)

        return tuple(result)
    else:
        return ()


_host_node = None


def buildAnnotationNode(provider, node, source_ref):
    if (
        python_version >= 0x370
        and provider.getParentModule().getFutureSpec().isFutureAnnotations()
    ):
        # Using global value for cache, to avoid creating it over and over,
        # avoiding the pylint: disable=global-statement
        global _host_node

        if _host_node is None:
            _host_node = ast.parse("x:1")

        _host_node.body[0].annotation = node

        r = compile(
            _host_node,
            "<annotations>",
            "exec",
            __future__.CO_FUTURE_ANNOTATIONS,
            dont_inherit=True,
        )

        # Using exec here, to compile the ast node tree back to string,
        # there is no accessible "ast.unparse", and this works as a hack
        # to convert our node to a string annotation, pylint: disable=exec-used
        m = {}
        exec(r, m)

        value = m["__annotations__"]["x"]

        if states.is_debug and python_version >= 0x390:
            # TODO: In Python3.9+, we should only use ast.unparse
            assert value == ast.unparse(node)

        return makeConstantRefNode(constant=value, source_ref=source_ref)

    return buildNode(provider, node, source_ref)


def makeModuleFrame(module, statements, source_ref):
    assert module.isCompiledPythonModule()

    if states.is_full_compat:
        co_name = "<module>"
    else:
        if module.isMainModule():
            co_name = "<module>"
        else:
            co_name = "<module %s>" % module.getFullName()

    return StatementsFrameModule(
        statements=tuple(statements),
        code_object=CodeObjectSpec(
            co_name=co_name,
            co_qualname=co_name,
            co_kind="Module",
            co_varnames=(),
            co_freevars=(),
            co_argcount=0,
            co_posonlyargcount=0,
            co_kwonlyargcount=0,
            co_has_starlist=False,
            co_has_stardict=False,
            co_filename=module.getRunTimeFilename(),
            co_lineno=source_ref.getLineNumber(),
            future_spec=module.getFutureSpec(),
        ),
        owner_code_name=module.getCodeName(),
        source_ref=source_ref,
    )


def buildStatementsNode(provider, nodes, source_ref):
    # We are not creating empty statement sequences.
    if nodes is None:
        return None

    # Build as list of statements, throw away empty ones, and remove useless
    # nesting.
    statements = buildNodeList(provider, nodes, source_ref, allow_none=True)
    statements = mergeStatements(statements)

    # We are not creating empty statement sequences. Might be empty, because
    # e.g. a global node generates not really a statement, or pass statements.
    if not statements:
        return None
    else:
        return StatementsSequence(statements=statements, source_ref=source_ref)


def buildFrameNode(provider, nodes, code_object, source_ref):
    # We are not creating empty statement sequences.
    if nodes is None:
        return None

    # Build as list of statements, throw away empty ones, and remove useless
    # nesting.
    statements = buildNodeList(provider, nodes, source_ref, allow_none=True)
    statements = mergeStatements(statements)

    # We are not creating empty statement sequences. Might be empty, because
    # e.g. a global node generates not really a statement, or pass statements.
    if not statements:
        return None

    if provider.isExpressionOutlineFunction():
        provider = provider.getParentVariableProvider()

    if provider.isExpressionFunctionBody():
        result = StatementsFrameFunction(
            statements=statements,
            code_object=code_object,
            owner_code_name=provider.getCodeName(),
            source_ref=source_ref,
        )
    elif provider.isExpressionClassBodyBase():
        result = StatementsFrameClass(
            statements=statements,
            code_object=code_object,
            owner_code_name=provider.getCodeName(),
            locals_scope=provider.getLocalsScope(),
            source_ref=source_ref,
        )
    elif provider.isExpressionGeneratorObjectBody():
        result = StatementsFrameGenerator(
            statements=statements,
            code_object=code_object,
            owner_code_name=provider.getCodeName(),
            source_ref=source_ref,
        )
    elif provider.isExpressionCoroutineObjectBody():
        result = StatementsFrameCoroutine(
            statements=statements,
            code_object=code_object,
            owner_code_name=provider.getCodeName(),
            source_ref=source_ref,
        )
    elif provider.isExpressionAsyncgenObjectBody():
        result = StatementsFrameAsyncgen(
            statements=statements,
            code_object=code_object,
            owner_code_name=provider.getCodeName(),
            source_ref=source_ref,
        )
    else:
        assert False, provider

    return result


def makeStatementsSequenceOrStatement(statements, source_ref):
    """Make a statement sequence, but only if more than one statement

    Useful for when we can unroll constructs already here, but are not sure if
    we actually did that. This avoids the branch or the pollution of doing it
    always.
    """

    if len(statements) > 1:
        return StatementsSequence(
            statements=mergeStatements(statements), source_ref=source_ref
        )
    else:
        return statements[0]


def makeStatementsSequence(statements, allow_none, source_ref):
    if allow_none:
        statements = tuple(
            statement for statement in statements if statement is not None
        )

    if statements:
        return StatementsSequence(
            statements=mergeStatements(statements, allow_none=allow_none),
            source_ref=source_ref,
        )
    else:
        return None


def makeStatementsSequenceFromStatement(statement):
    return StatementsSequence(
        statements=mergeStatements((statement,)),
        source_ref=statement.getSourceReference(),
    )


def makeStatementsSequenceFromStatements(*statements):
    assert statements
    assert None not in statements

    statements = mergeStatements(statements, allow_none=False)

    return StatementsSequence(
        statements=statements, source_ref=statements[0].getSourceReference()
    )


def makeDictCreationOrConstant2(keys, values, source_ref):
    # Create dictionary node. Tries to avoid it for constant values that are not
    # mutable. Keys are Python strings here.

    assert len(keys) == len(values)
    for value in values:
        if not value.isExpressionConstantRef():
            constant = False
            break
    else:
        constant = True

    # Note: This would happen in optimization instead, but lets just do it
    # immediately to save some time.
    if constant:
        # Unless told otherwise, create the dictionary in its full size, so
        # that no growing occurs and the constant becomes as similar as possible
        # before being marshaled.
        result = makeConstantRefNode(
            constant=createConstantDict(
                keys=keys, values=[value.getCompileTimeConstant() for value in values]
            ),
            user_provided=True,
            source_ref=source_ref,
        )
    else:
        result = makeExpressionMakeDict(
            pairs=makeKeyValuePairExpressionsFromKwArgs(zip(keys, values)),
            source_ref=source_ref,
        )

    if values:
        result.setCompatibleSourceReference(
            source_ref=values[-1].getCompatibleSourceReference()
        )

    return result


def getStatementsAppended(statement_sequence, statements):
    return makeStatementsSequence(
        statements=(statement_sequence, statements),
        allow_none=False,
        source_ref=statement_sequence.getSourceReference(),
    )


def getStatementsPrepended(statement_sequence, statements):
    return makeStatementsSequence(
        statements=(statements, statement_sequence),
        allow_none=False,
        source_ref=statement_sequence.getSourceReference(),
    )


def makeReraiseExceptionStatement(source_ref):
    return StatementReraiseException(source_ref=source_ref)


def mangleName(name, owner):
    """Mangle names with leading "__" for usage in a class owner.

    Notes: The is the private name handling for Python classes.
    """

    if not name.startswith("__") or name.endswith("__"):
        return name
    else:
        # The mangling of function variable names depends on being inside a
        # class.
        class_container = owner.getContainingClassDictCreation()

        if class_container is None:
            return name
        else:
            return "_%s%s" % (class_container.getName().lstrip("_"), name)


def makeCallNode(called, *args, **kwargs):
    source_ref = args[-1]

    if len(args) > 1:
        args = makeExpressionMakeTupleOrConstant(
            elements=args[:-1], user_provided=True, source_ref=source_ref
        )
    else:
        args = None

    if kwargs:
        kwargs = makeDictCreationOrConstant2(
            keys=tuple(kwargs.keys()),
            values=tuple(kwargs.values()),
            source_ref=source_ref,
        )
    else:
        kwargs = None

    return makeExpressionCall(
        called=called, args=args, kw=kwargs, source_ref=source_ref
    )


build_contexts = [None]


def pushBuildContext(value):
    build_contexts.append(value)


def popBuildContext():
    del build_contexts[-1]


def getBuildContext():
    return build_contexts[-1]


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#
# Copyright (c) 2001 - 2019 The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/CacheDir.py bee7caf9defd6e108fc2998a2520ddb36a967691 2019-12-17 02:07:09 bdeegan"

__doc__ = """
CacheDir support
"""

import hashlib
import json
import os
import stat
import sys

import SCons
import SCons.Action
import SCons.Warnings
from SCons.Util import PY3

cache_enabled = True
cache_debug = False
cache_force = False
cache_show = False
cache_readonly = False

def CacheRetrieveFunc(target, source, env):
    t = target[0]
    fs = t.fs
    cd = env.get_CacheDir()
    cd.requests += 1
    cachedir, cachefile = cd.cachepath(t)
    if not fs.exists(cachefile):
        cd.CacheDebug('CacheRetrieve(%s):  %s not in cache\n', t, cachefile)
        return 1
    cd.hits += 1
    cd.CacheDebug('CacheRetrieve(%s):  retrieving from %s\n', t, cachefile)
    if SCons.Action.execute_actions:
        if fs.islink(cachefile):
            fs.symlink(fs.readlink(cachefile), t.get_internal_path())
        else:
            env.copy_from_cache(cachefile, t.get_internal_path())
            try:
                os.utime(cachefile, None)
            except OSError:
                pass
        st = fs.stat(cachefile)
        fs.chmod(t.get_internal_path(), stat.S_IMODE(st[stat.ST_MODE]) | stat.S_IWRITE)
    return 0

def CacheRetrieveString(target, source, env):
    t = target[0]
    fs = t.fs
    cd = env.get_CacheDir()
    cachedir, cachefile = cd.cachepath(t)
    if t.fs.exists(cachefile):
        return "Retrieved `%s' from cache" % t.get_internal_path()
    return None

CacheRetrieve = SCons.Action.Action(CacheRetrieveFunc, CacheRetrieveString)

CacheRetrieveSilent = SCons.Action.Action(CacheRetrieveFunc, None)

def CachePushFunc(target, source, env):
    if cache_readonly:
        return

    t = target[0]
    if t.nocache:
        return
    fs = t.fs
    cd = env.get_CacheDir()
    cachedir, cachefile = cd.cachepath(t)
    if fs.exists(cachefile):
        # Don't bother copying it if it's already there.  Note that
        # usually this "shouldn't happen" because if the file already
        # existed in cache, we'd have retrieved the file from there,
        # not built it.  This can happen, though, in a race, if some
        # other person running the same build pushes their copy to
        # the cache after we decide we need to build it but before our
        # build completes.
        cd.CacheDebug('CachePush(%s):  %s already exists in cache\n', t, cachefile)
        return

    cd.CacheDebug('CachePush(%s):  pushing to %s\n', t, cachefile)

    tempfile = cachefile+'.tmp'+str(os.getpid())
    errfmt = "Unable to copy %s to cache. Cache file is %s"

    if not fs.isdir(cachedir):
        try:
            fs.makedirs(cachedir)
        except EnvironmentError:
            # We may have received an exception because another process
            # has beaten us creating the directory.
            if not fs.isdir(cachedir):
                msg = errfmt % (str(target), cachefile)
                raise SCons.Errors.SConsEnvironmentError(msg)

    try:
        if fs.islink(t.get_internal_path()):
            fs.symlink(fs.readlink(t.get_internal_path()), tempfile)
        else:
            fs.copy2(t.get_internal_path(), tempfile)
        fs.rename(tempfile, cachefile)
        st = fs.stat(t.get_internal_path())
        fs.chmod(cachefile, stat.S_IMODE(st[stat.ST_MODE]) | stat.S_IWRITE)
    except EnvironmentError:
        # It's possible someone else tried writing the file at the
        # same time we did, or else that there was some problem like
        # the CacheDir being on a separate file system that's full.
        # In any case, inability to push a file to cache doesn't affect
        # the correctness of the build, so just print a warning.
        msg = errfmt % (str(target), cachefile)
        SCons.Warnings.warn(SCons.Warnings.CacheWriteErrorWarning, msg)

CachePush = SCons.Action.Action(CachePushFunc, None)

# Nasty hack to cut down to one warning for each cachedir path that needs
# upgrading.
warned = dict()

class CacheDir(object):

    def __init__(self, path):
        """
        Initialize a CacheDir object.

        The cache configuration is stored in the object. It
        is read from the config file in the supplied path if
        one exists,  if not the config file is created and
        the default config is written, as well as saved in the object.
        """
        self.requests = 0
        self.hits = 0
        self.path = path
        self.current_cache_debug = None
        self.debugFP = None
        self.config = dict()
        if path is None:
            return

        if PY3:
            self._readconfig3(path)
        else:
            self._readconfig2(path)


    def _readconfig3(self, path):
        """
        Python3 version of reading the cache config.

        If directory or config file do not exist, create.  Take advantage
        of Py3 capability in os.makedirs() and in file open(): just try
        the operation and handle failure appropriately.

        Omit the check for old cache format, assume that's old enough
        there will be none of those left to worry about.

        :param path: path to the cache directory
        """
        config_file = os.path.join(path, 'config')
        try:
            os.makedirs(path, exist_ok=True)
        except FileExistsError:
            pass
        except OSError:
            msg = "Failed to create cache directory " + path
            raise SCons.Errors.SConsEnvironmentError(msg)

        try:
            with open(config_file, 'x') as config:
                self.config['prefix_len'] = 2
                try:
                    json.dump(self.config, config)
                except Exception:
                    msg = "Failed to write cache configuration for " + path
                    raise SCons.Errors.SConsEnvironmentError(msg)
        except FileExistsError:
            try:
                with open(config_file) as config:
                    self.config = json.load(config)
            except ValueError:
                msg = "Failed to read cache configuration for " + path
                raise SCons.Errors.SConsEnvironmentError(msg)


    def _readconfig2(self, path):
        """
        Python2 version of reading cache config.

        See if there is a config file in the cache directory. If there is,
        use it. If there isn't, and the directory exists and isn't empty,
        produce a warning. If the directory does not exist or is empty,
        write a config file.

        :param path: path to the cache directory
        """
        config_file = os.path.join(path, 'config')
        if not os.path.exists(config_file):
            # A note: There is a race hazard here if two processes start and
            # attempt to create the cache directory at the same time. However,
            # Python 2.x does not give you the option to do exclusive file
            # creation (not even the option to error on opening an existing
            # file for writing...). The ordering of events here is an attempt
            # to alleviate this, on the basis that it's a pretty unlikely
            # occurrence (would require two builds with a brand new cache
            # directory)
            if os.path.isdir(path) and any(f != "config" for f in os.listdir(path)):
                self.config['prefix_len'] = 1
                # When building the project I was testing this on, the warning
                # was output over 20 times. That seems excessive
                global warned
                if self.path not in warned:
                    msg = "Please upgrade your cache by running " +\
                          "scons-configure-cache.py " +  self.path
                    SCons.Warnings.warn(SCons.Warnings.CacheVersionWarning, msg)
                    warned[self.path] = True
            else:
                if not os.path.isdir(path):
                    try:
                        os.makedirs(path)
                    except OSError:
                        # If someone else is trying to create the directory at
                        # the same time as me, bad things will happen
                        msg = "Failed to create cache directory " + path
                        raise SCons.Errors.SConsEnvironmentError(msg)

                self.config['prefix_len'] = 2
                if not os.path.exists(config_file):
                    try:
                        with open(config_file, 'w') as config:
                            json.dump(self.config, config)
                    except Exception:
                        msg = "Failed to write cache configuration for " + path
                        raise SCons.Errors.SConsEnvironmentError(msg)
        else:
            try:
                with open(config_file) as config:
                    self.config = json.load(config)
            except ValueError:
                msg = "Failed to read cache configuration for " + path
                raise SCons.Errors.SConsEnvironmentError(msg)


    def CacheDebug(self, fmt, target, cachefile):
        if cache_debug != self.current_cache_debug:
            if cache_debug == '-':
                self.debugFP = sys.stdout
            elif cache_debug:
                self.debugFP = open(cache_debug, 'w')
            else:
                self.debugFP = None
            self.current_cache_debug = cache_debug
        if self.debugFP:
            self.debugFP.write(fmt % (target, os.path.split(cachefile)[1]))
            self.debugFP.write("requests: %d, hits: %d, misses: %d, hit rate: %.2f%%\n" %
                               (self.requests, self.hits, self.misses, self.hit_ratio))

    @property
    def hit_ratio(self):
        return (100.0 * self.hits / self.requests if self.requests > 0 else 100)

    @property
    def misses(self):
        return self.requests - self.hits

    def is_enabled(self):
        return cache_enabled and self.path is not None

    def is_readonly(self):
        return cache_readonly

    def cachepath(self, node):
        """
        """
        if not self.is_enabled():
            return None, None

        sig = node.get_cachedir_bsig()

        subdir = sig[:self.config['prefix_len']].upper()

        dir = os.path.join(self.path, subdir)
        return dir, os.path.join(dir, sig)

    def retrieve(self, node):
        """
        This method is called from multiple threads in a parallel build,
        so only do thread safe stuff here. Do thread unsafe stuff in
        built().

        Note that there's a special trick here with the execute flag
        (one that's not normally done for other actions).  Basically
        if the user requested a no_exec (-n) build, then
        SCons.Action.execute_actions is set to 0 and when any action
        is called, it does its showing but then just returns zero
        instead of actually calling the action execution operation.
        The problem for caching is that if the file does NOT exist in
        cache then the CacheRetrieveString won't return anything to
        show for the task, but the Action.__call__ won't call
        CacheRetrieveFunc; instead it just returns zero, which makes
        the code below think that the file *was* successfully
        retrieved from the cache, therefore it doesn't do any
        subsequent building.  However, the CacheRetrieveString didn't
        print anything because it didn't actually exist in the cache,
        and no more build actions will be performed, so the user just
        sees nothing.  The fix is to tell Action.__call__ to always
        execute the CacheRetrieveFunc and then have the latter
        explicitly check SCons.Action.execute_actions itself.
        """
        if not self.is_enabled():
            return False

        env = node.get_build_env()
        if cache_show:
            if CacheRetrieveSilent(node, [], env, execute=1) == 0:
                node.build(presub=0, execute=0)
                return True
        else:
            if CacheRetrieve(node, [], env, execute=1) == 0:
                return True

        return False

    def push(self, node):
        if self.is_readonly() or not self.is_enabled():
            return
        return CachePush(node, [], node.get_build_env())

    def push_if_forced(self, node):
        if cache_force:
            return self.push(node)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Reference counting tests for Python3.12 or higher.

These contain functions that do specific things, where we have a suspect that
references may be lost or corrupted. Executing them repeatedly and checking the
reference count is how they are used.

These are Python3.12 specific constructs, that will give a SyntaxError or not be
relevant on older versions.
"""

import os
import sys

# While we use that for comparison code, no need to compile that.
# nuitka-project: --nofollow-import-to=nuitka

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

from nuitka.tools.testing.Common import executeReferenceChecked


def simpleFunction1():
    class Parent[T]:
        value: T

    return Parent


def no_simpleFunction2():
    class Parent[T]:
        value: T

    class Child[T](Parent[T]):
        another_value: int

    return Child


# These need stderr to be wrapped.
tests_stderr = ()

# Disabled tests
tests_skipped = {}

result = executeReferenceChecked(
    prefix="simpleFunction",
    names=globals(),
    tests_skipped=tests_skipped,
    tests_stderr=tests_stderr,
    explain=False,
)

print("OK." if result else "FAIL.")
sys.exit(0 if result else 1)

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Common helper functions for specializing code."""

import contextlib

from nuitka.Constants import the_empty_unicode
from nuitka.tools.quality.auto_format.AutoFormat import (
    withFileOpenedAndAutoFormatted,
)
from nuitka.tools.release.Copyright import (
    attachLeadingComment,
    getCopyrightClaim,
)


def writeLine(output, *args):
    if not args:
        output.write("\n")
    elif len(args) == 1:
        output.write(args[0] + "\n")
    else:
        assert False, args


# Python2 dict methods:
python2_dict_methods = (  # we have full coverage for all methods
    "clear",
    "copy",
    "fromkeys",
    "get",
    "has_key",
    "items",
    "iteritems",
    "iterkeys",
    "itervalues",
    "keys",
    "pop",
    "popitem",
    "setdefault",
    "update",
    "values",
    "viewitems",
    "viewkeys",
    "viewvalues",
)

python3_dict_methods = (
    # see Python2 methods, these are only less
    "clear",
    "copy",
    "fromkeys",
    "get",
    "items",
    "keys",
    "pop",
    "popitem",
    "setdefault",
    "update",
    "values",
)

python2_str_methods = (  # we have full coverage for all methods
    "capitalize",
    "center",
    "count",
    "decode",
    "encode",
    "endswith",
    "expandtabs",
    "find",
    "format",
    "index",
    "isalnum",
    "isalpha",
    "isdigit",
    "islower",
    "isspace",
    "istitle",
    "isupper",
    "join",
    "ljust",
    "lower",
    "lstrip",
    "partition",
    "replace",
    "rfind",
    "rindex",
    "rjust",
    "rpartition",
    "rsplit",
    "rstrip",
    "split",
    "splitlines",
    "startswith",
    "strip",
    "swapcase",
    "title",
    "translate",
    "upper",
    "zfill",
)

python3_str_methods = (  # we have full coverage for all methods
    "capitalize",
    "casefold",
    "center",
    "count",
    "encode",
    "endswith",
    "expandtabs",
    "find",
    "format",
    "format_map",
    "index",
    "isalnum",
    "isalpha",
    "isascii",
    "isdecimal",
    "isdigit",
    "isidentifier",
    "islower",
    "isnumeric",
    "isprintable",
    "isspace",
    "istitle",
    "isupper",
    "join",
    "ljust",
    "lower",
    "lstrip",
    "maketrans",
    "partition",
    # TODO: Python3.9 or higher:
    # "removeprefix",
    # "removesuffix",
    "replace",
    "rfind",
    "rindex",
    "rjust",
    "rpartition",
    "rsplit",
    "rstrip",
    "split",
    "splitlines",
    "startswith",
    "strip",
    "swapcase",
    "title",
    "translate",
    "upper",
    "zfill",
)

python2_unicode_methods = (
    "capitalize",
    "center",
    "count",
    "decode",
    "encode",
    "endswith",
    "expandtabs",
    "find",
    "format",
    "index",
    "isalnum",
    "isalpha",
    "isdecimal",
    "isdigit",
    "islower",
    "isnumeric",
    "isspace",
    "istitle",
    "isupper",
    "join",
    "ljust",
    "lower",
    "lstrip",
    "partition",
    "replace",
    "rfind",
    "rindex",
    "rjust",
    "rpartition",
    "rsplit",
    "rstrip",
    "split",
    "splitlines",
    "startswith",
    "strip",
    "swapcase",
    "title",
    "translate",
    "upper",
    "zfill",
)

python3_bytes_methods = (
    "capitalize",
    "center",
    "count",
    "decode",
    "endswith",
    "expandtabs",
    "find",
    # static method
    # "fromhex",
    # TODO: Python3.5 or higher:
    # "hex",
    "index",
    "isalnum",
    "isalpha",
    "isdigit",
    "islower",
    "isspace",
    "istitle",
    "isupper",
    # TODO: Python3.7 or higher:
    # "ishex",
    # "isascii",
    "join",
    "ljust",
    "lower",
    "lstrip",
    # static method
    # "maketrans",
    "partition",
    # TODO: Python3.9 or higher:
    # "removeprefix",
    # "removesuffix",
    "replace",
    "rfind",
    "rindex",
    "rjust",
    "rpartition",
    "rsplit",
    "rstrip",
    "split",
    "splitlines",
    "startswith",
    "strip",
    "swapcase",
    "title",
    "translate",
    "upper",
    "zfill",
)

python2_list_methods = (
    "append",
    "count",
    "extend",
    "index",
    "insert",
    "pop",
    "remove",
    "reverse",
    "sort",
)

python3_list_methods = (
    "append",
    "clear",
    "copy",
    "count",
    "extend",
    "index",
    "insert",
    "pop",
    "remove",
    "reverse",
    "sort",
)

python2_type_methods = (
    # TODO: Totally incomplete
)

python3_type_methods = (
    # TODO: Totally incomplete
    "__prepare__",
)


def getMethodVariations(spec_module, shape_name, method_name, must_exist=False):
    spec_name = shape_name.split("_")[-1] + "_" + method_name + "_spec"
    spec = getattr(spec_module, spec_name, None)

    present = spec is not None

    if not present and must_exist:
        assert False, spec_name

    if present:
        if spec.isStarListSingleArg():
            required = 1

            arg_counts = tuple(range(required, required + 2))

            arg_names = (
                spec.getStarListArgumentName(),
                spec.getStarDictArgumentName(),
            )

            arg_name_mapping = {
                "list_args": spec.getStarListArgumentName(),
            }
        else:
            required = spec.getArgumentCount() - spec.getDefaultCount()

            arg_counts = tuple(range(required, spec.getArgumentCount() + 1))

            arg_names = spec.getParameterNames()
            arg_name_mapping = {}

        arg_tests = [
            (
                ""
                if arg_name
                in (spec.getStarListArgumentName(), spec.getStarDictArgumentName())
                else "is not None"
            )
            for arg_name in arg_names
        ]

        result_shape = spec.getTypeShape()

    else:
        arg_names = arg_name_mapping = arg_counts = arg_tests = result_shape = None

    return present, arg_names, arg_tests, arg_name_mapping, arg_counts, result_shape


def getSpecs(spec_module):
    for name in sorted(dir(spec_module)):
        if name.endswith("_spec"):
            yield name, getattr(spec_module, name)


def formatArgs(args, starting=True, finishing=True, quoted=False):
    result = []
    if args:
        if not starting:
            result.append(",")

        for arg in args:
            result.append('"%s"' % arg if quoted else arg)

            if arg is not args[-1] or not finishing:
                result.append(",")

    return "".join(result)


def check():
    if str is bytes:
        for method_name in python2_str_methods:
            assert hasattr("", method_name), method_name
        for method_name in python2_unicode_methods:
            assert hasattr(the_empty_unicode, method_name), method_name
        for method_name in python2_dict_methods:
            assert hasattr({}, method_name), method_name
    else:
        for method_name in python3_str_methods:
            assert hasattr("", method_name), method_name
        for method_name in python3_dict_methods:
            assert hasattr({}, method_name), method_name


check()


@contextlib.contextmanager
def withFileOpenedAndAutoFormattedWithClaim(filename, claim, ignore_errors=False):
    """Context manager for opening a file, auto-formatting it, and adding a copyright claim."""
    with withFileOpenedAndAutoFormatted(
        filename=filename,
        ignore_errors=ignore_errors,
    ) as output:
        yield output

    if claim:
        attachLeadingComment(
            filename=filename,
            effective_filename=filename,
            comments=getCopyrightClaim(filename=filename, claim=claim),
        )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Variable type for path Variables.

To be used whenever a user-specified path override setting should be allowed.

Arguments to PathVariable are:
  * *key* - name of this variable on the command line (e.g. "prefix")
  * *help* - help string for variable
  * *default* - default value for this variable
  * *validator* - [optional] validator for variable value.  Predefined are:

    * *PathAccept* - accepts any path setting; no validation
    * *PathIsDir* - path must be an existing directory
    * *PathIsDirCreate* - path must be a dir; will create
    * *PathIsFile* - path must be a file
    * *PathExists* - path must exist (any type) [default]

  The *validator* is a function that is called and which should return
  True or False to indicate if the path is valid.  The arguments
  to the validator function are: (*key*, *val*, *env*).  *key* is the
  name of the variable, *val* is the path specified for the variable,
  and *env* is the environment to which the Variables have been added.

Usage example::

    opts = Variables()
    opts.Add(
        PathVariable(
            'qtdir',
            help='where the root of Qt is installed',
            default=qtdir,
            validator=PathIsDir,
        )
    )
    opts.Add(
        PathVariable(
            'qt_includes',
            help='where the Qt includes are installed',
            default='$qtdir/includes',
            validator=PathIsDirCreate,
        )
    )
    opts.Add(
        PathVariable(
            'qt_libraries',
            help='where the Qt library is installed',
            default='$qtdir/lib',
        )
    )
"""

from __future__ import annotations

import os
import os.path
from typing import Callable

import SCons.Errors
import SCons.Util

__all__ = ['PathVariable',]

class _PathVariableClass:
    """Class implementing path variables.

    This class exists mainly to expose the validators without code having
    to import the names: they will appear as methods of ``PathVariable``,
    a statically created instance of this class, which is placed in
    the SConscript namespace.

    Instances are callable to produce a suitable variable tuple.
    """

    @staticmethod
    def PathAccept(key, val, env):
        """Validate path with no checking."""
        return

    @staticmethod
    def PathIsDir(key, val, env):
        """Validate path is a directory."""
        if os.path.isdir(val):
            return
        if os.path.isfile(val):
            msg = f'Directory path for variable {key!r} is a file: {val}'
        else:
            msg = f'Directory path for variable {key!r} does not exist: {val}'
        raise SCons.Errors.UserError(msg)

    @staticmethod
    def PathIsDirCreate(key, val, env):
        """Validate path is a directory, creating if needed."""
        if os.path.isdir(val):
            return
        try:
            os.makedirs(val, exist_ok=True)
        except FileExistsError as exc:
            msg = f'Path for variable {key!r} is a file, not a directory: {val}'
            raise SCons.Errors.UserError(msg) from exc
        except (PermissionError, OSError) as exc:
            msg = f'Path for variable {key!r} could not be created: {val}'
            raise SCons.Errors.UserError(msg) from exc

    @staticmethod
    def PathIsFile(key, val, env):
        """Validate path is a file."""
        if not os.path.isfile(val):
            if os.path.isdir(val):
                msg = f'File path for variable {key!r} is a directory: {val}'
            else:
                msg = f'File path for variable {key!r} does not exist: {val}'
            raise SCons.Errors.UserError(msg)

    @staticmethod
    def PathExists(key, val, env):
        """Validate path exists."""
        if not os.path.exists(val):
            msg = f'Path for variable {key!r} does not exist: {val}'
            raise SCons.Errors.UserError(msg)

    # lint: W0622: Redefining built-in 'help' (redefined-builtin)
    def __call__(
        self, key, help, default, validator = None
    ):
        """Return a tuple describing a path list SCons Variable.

        The input parameters describe a 'path list' variable. Returns
        a tuple with the correct converter and validator appended. The
        result is usable for input to :meth:`Add`.

        The *default* parameter specifies the default path to use if the
        user does not specify an override with this variable.

        *validator* is a validator, see this file for examples
        """
        if validator is None:
            validator = self.PathExists

        if SCons.Util.is_List(key) or SCons.Util.is_Tuple(key):
            helpmsg = f'{help} ( /path/to/{key[0]} )'
        else:
            helpmsg = f'{help} ( /path/to/{key} )'
        return key, helpmsg, default, validator, None


PathVariable = _PathVariableClass()

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
This module handles the extraction of build configuration from UV Build projects.
"""

import os
import sys

from nuitka.importing.Importing import addMainScriptDirectory
from nuitka.utils.Execution import executeProcess
from nuitka.utils.FileOperations import withTemporaryDirectory
from nuitka.utils.Json import loadJsonFromFilename

from .BuildPackageCommon import (
    reportBuildError,
    setProjectExpectedDataFiles,
    setProjectName,
)


def _maybeVirtualEnv(path):
    """
    Check if a directory might be a virtualenv.
    """
    # Windows
    if os.path.exists(os.path.join(path, "Scripts", "python.exe")):
        return True

    # Unix
    if os.path.exists(os.path.join(path, "bin", "python")):
        return True

    return False


def getUvBuildConfiguration(logger, pyproject_data):
    # Necessarily a lot of details in the configuration
    # pylint: disable=too-many-locals
    """
    Get the build configuration from a UV Build project.
    """
    with withTemporaryDirectory("nuitka-project-dump") as temp_dir:
        dump_filename = os.path.join(temp_dir, "build_config.json")

        # Locate the script "tools/general/extract_uv_config/__main__.py" relative to this module.
        # We assume this module is in "nuitka/options/UvBuild.py"
        script_filename = os.path.join(
            os.path.dirname(__file__),
            "..",
            "tools",
            "general",
            "extract_uv_config",
        )

        command = (
            sys.executable,
            script_filename,
            dump_filename,
        )

        stdout, stderr, exit_code = executeProcess(
            command,
            stdin=False,
        )

        if exit_code != 0:
            reportBuildError(logger, "uv_build", command, stdout, stderr)

        if not os.path.exists(dump_filename):
            logger.sysexit(
                "Error, 'uv_build' configuration extraction produced no output file."
            )

        config = loadJsonFromFilename(dump_filename)

        arguments = config.get("arguments", [])

        # Parse tool.uv configuration
        uv_config = dict(pyproject_data.get("tool", {}).get("uv", {}))

        module_root = uv_config.pop("module-root", None)
        if module_root is not None:
            if module_root != "":
                addMainScriptDirectory(os.path.abspath(module_root))
        elif os.path.exists("src"):
            addMainScriptDirectory(os.path.abspath("src"))

        if uv_config:
            for unhandled_key in uv_config:
                logger.warning(
                    """\
Unhandled UV config key '%s' in [tool.uv] of the 'pyproject.toml', we might \
have to ignore list or handle it: %s"""
                    % (unhandled_key, uv_config[unhandled_key])
                )

        detected_packages = config.get("packages", [])

        # Add --include-package-data for detected packages
        # Only if not already specified by other means (though duplicate args are usually fine/merged)
        for package_name in detected_packages:
            arguments.append("--include-package-data=%s" % package_name)
            arguments.append("--include-package=%s" % package_name)

        setProjectName(config.get("project_name"))
        setProjectExpectedDataFiles(config.get("data_files", []))

        # TODO: Check against IncludedDataFiles set once that is considered complete

        return arguments


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Handling of syntax errors.

Format SyntaxError/IndentationError exception for output, as well as
raise it for the given source code reference.
"""

from nuitka.PythonVersions import python_version


def formatOutput(e):
    if len(e.args) > 1:
        if len(e.args[1]) == 4:
            reason, (filename, lineno, colno, message) = e.args
        else:
            reason, (filename, lineno, colno, message, _lineno2, _colno2) = e.args

        if message is None and colno is not None:
            colno = None

        if lineno is not None and lineno == 0:
            lineno = 1
    else:
        (reason,) = e.args

        filename = None
        lineno = None
        colno = None
        message = None

    # On CPython3.4 at least, this attribute appears to override reason for
    # SyntaxErrors at least.
    if hasattr(e, "msg"):
        reason = e.msg

    if colno is not None and (
        not e.__class__ is IndentationError or python_version < 0x390
    ):
        colno = colno - len(message) + len(message.lstrip())

        return """\
  File "%s", line %d
    %s
    %s^
%s: %s""" % (
            filename,
            lineno,
            message.strip(),
            " " * (colno - 1) if colno is not None else "",
            e.__class__.__name__,
            reason,
        )
    elif message is not None:
        return """\
  File "%s", line %d
    %s
%s: %s""" % (
            filename,
            lineno,
            message.strip(),
            e.__class__.__name__,
            reason,
        )
    elif filename is not None:
        return """\
  File "%s", line %s
%s: %s""" % (
            filename,
            lineno,
            e.__class__.__name__,
            reason,
        )
    else:
        return """\
%s: %s""" % (
            e.__class__.__name__,
            reason,
        )


def raiseSyntaxError(reason, source_ref, display_file=True, display_line=True):
    col_offset = source_ref.getColumnNumber()

    def readSource():
        # Cyclic dependency.
        from .SourceHandling import readSourceLine

        return readSourceLine(source_ref)

    # Special case being asked to display.
    if display_file and display_line:
        source_line = readSource()

        raise SyntaxError(
            reason,
            (
                source_ref.getFilename(),
                source_ref.getLineNumber(),
                col_offset,
                source_line,
            ),
        )

    # Special case given source reference.
    if source_ref is not None:
        if display_line:
            source_line = readSource()
        else:
            source_line = None

        exc = SyntaxError(
            reason,
            (
                source_ref.getFilename(),
                source_ref.getLineNumber(),
                col_offset,
                source_line,
            ),
        )

    exc = SyntaxError(reason, (None, None, None, None))

    exc.generated_by_nuitka = True

    raise exc


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
